"""Tests for runway.cfngin entry point."""
# pylint: disable=no-self-use,protected-access
import os
import shutil

import pytest
from mock import MagicMock, patch

from runway.cfngin import CFNgin
from runway.context import Context
from runway.util import AWS_ENV_VARS


def copy_fixture(src, dest):
    """Wrap shutil.copy to backport use with Path objects."""
    return shutil.copy(str(src.absolute()), str(dest.absolute()))


def copy_basic_fixtures(cfngin_fixtures, tmp_path):
    """Copy the basic env file and config file to a tmp_path."""
    copy_fixture(src=cfngin_fixtures / 'envs' / 'basic.env',
                 dest=tmp_path / 'test-us-east-1.env')
    copy_fixture(src=cfngin_fixtures / 'configs' / 'basic.yml',
                 dest=tmp_path / 'basic.yml')


def get_env_creds():
    """Return AWS creds from the environment."""
    return {name: os.environ.get(name) for name in AWS_ENV_VARS if os.environ.get(name)}


class TestCFNgin(object):
    """Test runway.cfngin.CFNgin."""

    @staticmethod
    def configure_mock_action_instance(mock_action):
        """Configure a mock action."""
        mock_instance = MagicMock(return_value=None)
        mock_action.return_value = mock_instance
        mock_instance.execute = MagicMock()
        return mock_instance

    @staticmethod
    def get_context(name='test', region='us-east-1'):
        """Create a basic Runway context object."""
        return Context(env_name=name,
                       env_region=region,
                       env_root=os.getcwd())

    def test_env_file(self, tmp_path):
        """Test that the correct env file is selected."""
        test_env = tmp_path / 'test.env'
        # python2 Path.write_text requires unicode
        test_env.write_text(u'test_value: test')

        # support python < 3.6
        result = CFNgin(ctx=self.get_context(), sys_path=str(tmp_path))
        assert result.env_file.test_value == 'test'

        test_us_east_1 = tmp_path / 'test-us-east-1.env'
        # python2 Path.write_text requires unicode
        test_us_east_1.write_text(u'test_value: test-us-east-1')

        test_us_west_2 = tmp_path / 'test-us-west-2.env'
        # python2 Path.write_text requires unicode
        test_us_west_2.write_text(u'test_value: test-us-west-2')

        lab_ca_central_1 = tmp_path / 'lab-ca-central-1.env'
        # python2 Path.write_text requires unicode
        lab_ca_central_1.write_text(u'test_value: lab-ca-central-1')

        # support python < 3.6
        result = CFNgin(ctx=self.get_context(), sys_path=str(tmp_path))
        assert result.env_file.test_value == 'test-us-east-1'

        result = CFNgin(ctx=self.get_context(region='us-west-2'),
                        sys_path=str(tmp_path))  # support python < 3.6
        assert result.env_file.test_value == 'test-us-west-2'

        result = CFNgin(ctx=self.get_context(name='lab',
                                             region='ca-central-1'),
                        sys_path=str(tmp_path))  # support python < 3.6
        assert result.env_file.test_value == 'lab-ca-central-1'

    @patch('runway.cfngin.actions.build.Action')
    def test_deploy(self, mock_action, cfngin_fixtures, tmp_path):
        """Test deploy with two files & class init."""
        mock_instance = self.configure_mock_action_instance(mock_action)
        copy_basic_fixtures(cfngin_fixtures, tmp_path)
        copy_fixture(src=cfngin_fixtures / 'configs' / 'basic.yml',
                     dest=tmp_path / 'basic2.yml')

        context = self.get_context()
        context.env_vars['CI'] = '1'

        cfngin = CFNgin(ctx=context,
                        parameters={'test_param': 'test-param-value'},
                        sys_path=str(tmp_path))  # support python < 3.6
        cfngin.deploy()

        assert get_env_creds() == cfngin._aws_credential_backup, \
            'env vars should be reverted upon completion'
        assert cfngin.concurrency == 0
        assert not cfngin.interactive
        assert cfngin.parameters.bucket_name == 'cfngin-bucket'
        assert cfngin.parameters.environment == 'test'
        assert cfngin.parameters.namespace == 'test-namespace'
        assert cfngin.parameters.region == 'us-east-1'
        assert cfngin.parameters.test_key == 'test_value'
        assert cfngin.parameters.test_param == 'test-param-value'
        assert cfngin.recreate_failed
        assert cfngin.region == 'us-east-1'
        assert cfngin.sys_path == str(tmp_path)
        assert not cfngin.tail

        assert mock_action.call_count == 2
        mock_instance.execute.has_calls([{'concurrency': 0,
                                          'tail': False},
                                         {'concurrency': 0,
                                          'tail': False}])

    @patch('runway.cfngin.actions.destroy.Action')
    def test_destroy(self, mock_action, cfngin_fixtures, tmp_path):
        """Test destroy."""
        mock_instance = self.configure_mock_action_instance(mock_action)
        copy_basic_fixtures(cfngin_fixtures, tmp_path)

        # support python < 3.6
        cfngin = CFNgin(ctx=self.get_context(), sys_path=str(tmp_path))
        cfngin.destroy()

        assert get_env_creds() == cfngin._aws_credential_backup, \
            'env vars should be reverted upon completion'
        mock_action.assert_called_once()
        mock_instance.execute.assert_called_once_with(concurrency=0,
                                                      force=True,
                                                      tail=False)

    def test_load(self, cfngin_fixtures, tmp_path):
        """Test load."""
        copy_basic_fixtures(cfngin_fixtures, tmp_path)
        # support python < 3.6
        cfngin = CFNgin(ctx=self.get_context(), sys_path=str(tmp_path))
        result = cfngin.load(str(tmp_path / 'basic.yml'))  # support python < 3.6

        assert not result.bucket_name
        assert result.namespace == 'test-namespace'
        assert len(result.get_stacks()) == 1
        assert result.get_stacks()[0].name == 'test-stack'

    def test_load_cfn_template(self, caplog, tmp_path):
        """Test load a CFN template."""
        cfn_template = tmp_path / 'template.yml'
        cfn_template.write_text(u'test_key: !Ref something')
        cfngin = CFNgin(ctx=self.get_context(), sys_path=str(tmp_path))

        caplog.set_level('ERROR', logger='runway.cfngin')

        with pytest.raises(SystemExit):
            cfngin.load(str(cfn_template))  # support python < 3.6

        assert 'appears to be a CloudFormation template' in caplog.text

    @patch('runway.cfngin.actions.diff.Action')
    def test_plan(self, mock_action, cfngin_fixtures, tmp_path):
        """Test plan."""
        mock_instance = self.configure_mock_action_instance(mock_action)
        copy_basic_fixtures(cfngin_fixtures, tmp_path)

        # support python < 3.6
        cfngin = CFNgin(ctx=self.get_context(), sys_path=str(tmp_path))
        cfngin.plan()

        assert get_env_creds() == cfngin._aws_credential_backup, \
            'env vars should be reverted upon completion'
        mock_action.assert_called_once()
        mock_instance.execute.assert_called_once_with()

    def test_should_skip(self, cfngin_fixtures, tmp_path):
        """Test should_skip."""
        # support python < 3.6
        cfngin = CFNgin(ctx=self.get_context(), sys_path=str(tmp_path))
        del cfngin.env_file  # clear cached value and force load

        assert cfngin.should_skip()
        del cfngin.env_file  # clear cached value and force load
        assert not cfngin.should_skip(force=True)  # does not repopulate env_file

        copy_basic_fixtures(cfngin_fixtures, tmp_path)
        assert not cfngin.should_skip()
        del cfngin.env_file  # clear cached value and force load
        assert not cfngin.should_skip(force=True)  # does not repopulate env_file

        env_region_file = tmp_path / 'test-us-east-1.env'
        env_file = tmp_path / 'test.env'
        copy_fixture(env_region_file, env_file)
        env_region_file.unlink()
        assert not cfngin.should_skip()
        del cfngin.env_file  # clear cached value and force load
        assert not cfngin.should_skip(force=True)  # does not repopulate env_file

    def test_find_config_files(self, tmp_path):
        """Test find_config_files."""
        bad_path = tmp_path / 'bad_path'
        bad_path.mkdir()
        # tmp_path.stem

        good_config_paths = [
            tmp_path / '_t3sT.yaml',
            tmp_path / '_t3sT.yml',
            tmp_path / '01-config.yaml',
            tmp_path / '01-config.yml',
            tmp_path / 'TeSt_02.yaml',
            tmp_path / 'TeSt_02.yml',
            tmp_path / 'test.config.yaml',
            tmp_path / 'test.config.yml'
        ]
        bad_config_paths = [
            tmp_path / '.anything.yaml',
            tmp_path / '.gitlab-ci.yml',
            tmp_path / 'docker-compose.yml',
            bad_path / '00-invalid.yml'
        ]

        for config_path in good_config_paths + bad_config_paths:
            # python2 Path.write_text requires unicode
            config_path.write_text(u'')

        # support python < 3.6
        result = CFNgin.find_config_files(sys_path=str(tmp_path))
        expected = sorted([str(config_path)
                           for config_path in good_config_paths])
        assert result == expected

        config_01 = tmp_path / '01-config.yml'
        result = CFNgin.find_config_files(
            sys_path=str(config_01)  # support python < 3.6
        )
        assert result == [str(config_01)]  # support python < 3.6

        result = CFNgin.find_config_files()
        assert not result



"""Provides a sublass of unittest.TestCase for testing blueprints."""
import difflib
import json
import os.path
import unittest
from glob import glob

from runway.util import load_object_from_string
from runway.variables import Variable

from ..config import parse as parse_config
from ..context import Context


def diff(first, second):
    """Human readable differ."""
    return '\n'.join(
        list(
            difflib.Differ().compare(
                first.splitlines(),
                second.splitlines()
            )
        )
    )


class BlueprintTestCase(unittest.TestCase):
    """Extends the functionality of unittest.TestCase for testing blueprints."""

    OUTPUT_PATH = "tests/fixtures/blueprints"

    def assertRenderedBlueprint(self, blueprint):  # noqa: N802 pylint: disable=invalid-name
        """Test that the rendered blueprint json matches the expected result.

        Result files are to be stored in the repo as
        ``test/fixtures/blueprints/${blueprint.name}.json``.

        """
        expected_output = "%s/%s.json" % (self.OUTPUT_PATH, blueprint.name)

        rendered_dict = blueprint.template.to_dict()
        rendered_text = json.dumps(rendered_dict, indent=4, sort_keys=True)

        with open(expected_output + "-result", "w") as expected_output_file:
            expected_output_file.write(rendered_text)

        with open(expected_output) as expected_output_file:
            expected_dict = json.loads(expected_output_file.read())
            expected_text = json.dumps(expected_dict, indent=4, sort_keys=True)

        self.assertEqual(rendered_dict, expected_dict,
                         diff(rendered_text, expected_text))


class YamlDirTestGenerator(object):
    """Generate blueprint tests from yaml config files.

    This class creates blueprint tests from yaml files with a syntax similar
    to CFNgin configuration syntax. For example::

        namespace: test
        stacks:
            - name: test_sample
            class_path: blueprints.test.Sample
            variables:
                var1: value1

    Will create a test for the specified blueprint, passing that variable as
    part of the test.

    The test will generate a ``.json`` file for this blueprint, and compare it
    with the stored result.

    By default, the generator looks for files named ``test_*.yaml`` in its same
    directory. In order to use it, subclass it in a directory containing such
    tests, and name the class with a pattern that will include it in nosetests'
    tests (for example, TestGenerator).

    The subclass may override some ``@property`` definitions:

    **base_class**
      By default, the generated tests are subclasses or
      :class:`runway.cfngin.blueprints.testutil.BlueprintTestCase`. In order
      to change this, set this property to the desired base class.

    **yaml_dirs:**
      By default, the directory where the generator is subclassed is searched
      for test files. Override this array for specifying more directories.
      These must be relative to the directory in which the subclass lives in.
      Globs may be used. Default: ``['.']``.
      Example override: ``['.', 'tests/*/']``

    **yaml_filename:**
      By default, the generator looks for files named ``test_*.yaml``.
      Use this to change this pattern. Globs may be used.

    """

    def __init__(self):
        """Instantiate class."""
        self.classdir = os.path.relpath(
            self.__class__.__module__.replace('.', '/'))
        if not os.path.isdir(self.classdir):
            self.classdir = os.path.dirname(self.classdir)

    # These properties can be overridden from the test generator subclass.
    @property
    def base_class(self):
        """Return the baseclass."""
        return BlueprintTestCase

    @property
    def yaml_dirs(self):
        """Yaml directories."""
        return ['.']

    @property
    def yaml_filename(self):
        """Yaml filename."""
        return 'test_*.yaml'

    # pylint incorrectly detects this
    def test_generator(self):  # pylint: disable=no-self-use
        """Test generator."""
        # Search for tests in given paths
        configs = []
        for directory in self.yaml_dirs:
            configs.extend(
                glob('%s/%s/%s' % (self.classdir, directory,
                                   self.yaml_filename)))

        class ConfigTest(self.base_class):
            """Config test."""

            def __init__(self, config, stack, filepath):  # pylint: disable=super-init-not-called
                """Instantiate class."""
                self.config = config
                self.stack = stack
                self.description = "%s (%s)" % (stack.name, filepath)

            def __call__(self):  # pylint: disable=arguments-differ
                """Run when the class instance is called directly."""
                # Use the context property of the baseclass, if present.
                # If not, default to a basic context.
                try:
                    ctx = self.context
                except AttributeError:
                    ctx = Context(config=self.config,
                                  environment={'environment': 'test'})

                configvars = self.stack.variables or {}
                variables = [Variable(k, v, 'cfngin')
                             for k, v in configvars.iteritems()]

                blueprint_class = load_object_from_string(
                    self.stack.class_path)
                blueprint = blueprint_class(self.stack.name, ctx)
                blueprint.resolve_variables(variables or [])
                blueprint.setup_parameters()
                blueprint.create_template()
                self.assertRenderedBlueprint(blueprint)

            def assertEqual(self, first, second, msg=None):  # noqa pylint: disable=invalid-name
                """Test that first and second are equal.

                If the values do not compare equal, the test will fail.

                """
                assert first == second, msg

        for config_file in configs:
            with open(config_file) as test:
                config = parse_config(test.read())
                config.validate()

                for stack in config.stacks:  # pylint: disable=not-an-iterable
                    # Nosetests supports "test generators", which allows us to
                    # yield a callable object which will be wrapped as a test
                    # case.
                    #
                    # http://nose.readthedocs.io/en/latest/writing_tests.html#test-generators
                    yield ConfigTest(config, stack, filepath=config_file)



"""CFNgin base action."""
import logging
import os
import sys
import threading

import botocore.exceptions

from ..dag import ThreadedWalker, UnlimitedSemaphore, walk
from ..exceptions import PlanFailed
from ..plan import Graph, Plan, Step, merge_graphs
from ..status import COMPLETE
from ..util import ensure_s3_bucket, get_s3_endpoint, stack_template_key_name

LOGGER = logging.getLogger(__name__)

# After submitting a stack update/create, this controls how long we'll wait
# between calls to DescribeStacks to check on it's status. Most stack updates
# take at least a couple minutes, so 30 seconds is pretty reasonable and inline
# with the suggested value in
# https://github.com/boto/botocore/blob/1.6.1/botocore/data/cloudformation/2010-05-15/waiters-2.json#L22
#
# This can be controlled via an environment variable, mostly for testing.
STACK_POLL_TIME = int(os.environ.get("CFNGIN_STACK_POLL_TIME", 30))


def build_walker(concurrency):
    """Return a function for waling a graph.

    Passed to :class:`runway.cfngin.plan.Plan` for walking the graph.

    If concurrency is 1 (no parallelism) this will return a simple topological
    walker that doesn't use any multithreading.

    If concurrency is 0, this will return a walker that will walk the graph as
    fast as the graph topology allows.

    If concurrency is greater than 1, it will return a walker that will only
    execute a maximum of concurrency steps at any given time.

    Args:
        concurrency (int): Number of threads to use while walking.

    Returns:
        Callable[..., Any]: Function to walk a :class:`runway.cfngin.dag.DAG`.

    """
    if concurrency == 1:
        return walk

    semaphore = UnlimitedSemaphore()
    if concurrency > 1:
        semaphore = threading.Semaphore(concurrency)

    return ThreadedWalker(semaphore).walk


def stack_template_url(bucket_name, blueprint, endpoint):
    """Produce an s3 url for a given blueprint.

    Args:
        bucket_name (str): The name of the S3 bucket where the resulting
            templates are stored.
        blueprint (:class:`runway.cfngin.blueprints.base.Blueprint`): The
            blueprint object to create the URL to.
        endpoint (str): The s3 endpoint used for the bucket.

    Returns:
        str: S3 URL.

    """
    key_name = stack_template_key_name(blueprint)
    return "%s/%s/%s" % (endpoint, bucket_name, key_name)


class BaseAction(object):
    """Actions perform the actual work of each Command.

    Each action is tied to a :class:`runway.cfngin.commands.stacker.base.BaseCommand`,
    and is responsible for building the :class:`runway.cfngin.plan.Plan` that
    will be executed to perform that command.

    Attributes:
        DESCRIPTION (str): Description used when creating a plan for an action.
        bucket_name (str): S3 bucket used by the action.
        bucket_region (str): AWS region where S3 bucket is located.
        cancel (threading.Event): Cancel handler.
        context (:class:`runway.cfngin.context.Context`): The context
            for the current run.
        provider_builder (Optional[BaseProviderBuilder]):
            An object that will build a provider that will be interacted
            with in order to perform the necessary actions.
        s3_conn (boto3.client.Client): Boto3 S3 client.

    """

    DESCRIPTION = 'Base action'

    def __init__(self, context, provider_builder=None, cancel=None):
        """Instantiate class.

        Args:
            context (:class:`runway.cfngin.context.Context`): The context
                for the current run.
            provider_builder (Optional[:class:`BaseProviderBuilder`]):
                An object that will build a provider that will be interacted
                with in order to perform the necessary actions.
            cancel (threading.Event): Cancel handler.

        """
        self.context = context
        self.provider_builder = provider_builder
        self.bucket_name = context.bucket_name
        self.cancel = cancel or threading.Event()
        self.bucket_region = context.config.cfngin_bucket_region
        if not self.bucket_region and provider_builder:
            self.bucket_region = provider_builder.region
        self.s3_conn = self.context.s3_conn

    @property
    def _stack_action(self):
        """Run against a step."""
        raise NotImplementedError

    @property
    def provider(self):
        """Return a generic provider using the default region.

        Used for running things like hooks.

        Returns:
            :class:`runway.cfngin.providers.base.BaseProvider`

        """
        return self.provider_builder.build()

    def build_provider(self, stack):
        """Build a :class:`runway.cfngin.providers.base.BaseProvider`.

        Args:
            stack (:class:`runway.cfngin.stack.Stack`): Stack the action will
                be executed on.

        Returns:
            :class:`runway.cfngin.providers.base.BaseProvider`: Suitable for
            operating on the given :class:`runway.cfngin.stack.Stack`.

        """
        return self.provider_builder.build(region=stack.region,
                                           profile=stack.profile)

    def ensure_cfn_bucket(self):
        """CloudFormation bucket where templates will be stored."""
        if self.bucket_name:
            ensure_s3_bucket(self.s3_conn,
                             self.bucket_name,
                             self.bucket_region)

    def execute(self, **kwargs):
        """Run the action with pre and post steps."""
        try:
            self.pre_run(**kwargs)
            self.run(**kwargs)
            self.post_run(**kwargs)
        except PlanFailed as err:
            LOGGER.error(str(err))
            sys.exit(1)

    def pre_run(self, **kwargs):
        """Perform steps before running the action."""

    def post_run(self, **kwargs):
        """Perform steps after running the action."""

    def run(self, **kwargs):
        """Abstract method for running the action."""
        raise NotImplementedError("Subclass must implement \"run\" method")

    def s3_stack_push(self, blueprint, force=False):
        """Push the rendered blueprint's template to S3.

        Verifies that the template doesn't already exist in S3 before
        pushing.

        Returns:
            str: URL to the template in S3.

        """
        key_name = stack_template_key_name(blueprint)
        template_url = self.stack_template_url(blueprint)
        try:
            template_exists = self.s3_conn.head_object(
                Bucket=self.bucket_name, Key=key_name) is not None
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'] == '404':
                template_exists = False
            else:
                raise

        if template_exists and not force:
            LOGGER.debug("Cloudformation template %s already exists.",
                         template_url)
            return template_url
        self.s3_conn.put_object(Bucket=self.bucket_name,
                                Key=key_name,
                                Body=blueprint.rendered,
                                ServerSideEncryption='AES256',
                                ACL='bucket-owner-full-control')
        LOGGER.debug("Blueprint %s pushed to %s.", blueprint.name,
                     template_url)
        return template_url

    def stack_template_url(self, blueprint):
        """S3 URL for CloudFormation template object.

        Returns:
            str

        """
        return stack_template_url(
            self.bucket_name, blueprint, get_s3_endpoint(self.s3_conn)
        )

    def _generate_plan(self, tail=False, reverse=False,
                       require_unlocked=True,
                       include_persistent_graph=False):
        """Create a plan for this action.

        Args:
            tail (Union[bool, Callable]): An optional function to call
                to tail the stack progress.
            reverse (bool): If True, execute the graph in reverse (useful for
                destroy actions).
            require_unlocked (bool): If the persistent graph is locked, an
                error is raised.
            include_persistent_graph (bool): Include the persistent graph
                in the :class:`runway.cfngin.plan.Plan` (if there is one).
                This will handle basic merging of the local and persistent
                graphs if an action does not require more complex logic.

        Returns:
            :class:`runway.cfngin.plan.Plan`: The resulting plan object

        """
        tail = self._tail_stack if tail else None

        def target_fn(*_args, **_kwargs):
            return COMPLETE

        steps = [
            Step(stack, fn=self._stack_action, watch_func=tail)
            for stack in self.context.get_stacks()]

        steps += [
            Step(target, fn=target_fn)
            for target in self.context.get_targets()]

        graph = Graph.from_steps(steps)

        if include_persistent_graph and self.context.persistent_graph:
            persist_steps = Step.from_persistent_graph(
                self.context.persistent_graph.to_dict(),
                self.context,
                fn=self._stack_action,
                watch_func=tail
            )
            persist_graph = Graph.from_steps(persist_steps)
            graph = merge_graphs(graph, persist_graph)

        return Plan(
            context=self.context,
            description=self.DESCRIPTION,
            graph=graph,
            reverse=reverse,
            require_unlocked=require_unlocked)

    def _tail_stack(self, stack, cancel, retries=0, **kwargs):
        """Tail a stack's event stream."""
        provider = self.build_provider(stack)
        return provider.tail_stack(stack, cancel, retries, **kwargs)



"""Retrieve the ID of the Cognito User Pool."""
import logging

from typing import Any, Dict, Optional  # pylint: disable=unused-import

from runway.cfngin.providers.base import BaseProvider  # pylint: disable=unused-import
from runway.cfngin.context import Context  # noqa pylint: disable=unused-import

LOGGER = logging.getLogger(__name__)


def get(context,  # pylint: disable=unused-argument
        provider,  # pylint: disable=unused-argument
        **kwargs
       ):  # noqa: E124
    # type: (Context, BaseProvider, Optional[Dict[str, Any]]) -> Dict
    """Retrieve the ID of the Cognito User Pool.

    The User Pool can either be supplied via an ARN or by being generated.
    If the user has supplied an ARN that utilize that, otherwise retrieve
    the generated id. Used in multiple pre_hooks for Auth@Edge.

    Args:
        context (:class:`runway.cfngin.context.Context`): The context
            instance.
        provider (:class:`runway.cfngin.providers.base.BaseProvider`):
            The provider instance

    Keyword Args:
        user_pool_arn (str): The ARN of the supplied User pool
        created_user_pool_id (str): The ID of the created Cognito User Pool
    """
    context_dict = {'id': ''}

    # Favor a specific arn over a created one
    if kwargs['user_pool_arn']:
        context_dict['id'] = kwargs['user_pool_arn'].split('/')[-1:][0]
    elif kwargs['created_user_pool_id']:
        context_dict['id'] = kwargs['created_user_pool_id']

    return context_dict



"""Provider base class."""
# pylint: disable=no-self-use,too-few-public-methods,unused-argument


def not_implemented(method):
    """Wrap NotImplimentedError with a formatted message."""
    raise NotImplementedError("Provider does not support '%s' "
                              "method." % method)


class BaseProviderBuilder(object):
    """ProviderBuilder base class."""

    def build(self, region=None):
        """Abstract method."""
        not_implemented("build")


class BaseProvider(object):
    """Provider base class."""

    def get_stack(self, stack_name, *args, **kwargs):
        """Abstract method."""
        not_implemented("get_stack")

    def create_stack(self, *args, **kwargs):
        """Abstract method."""
        not_implemented("create_stack")

    def update_stack(self, *args, **kwargs):
        """Abstract method."""
        not_implemented("update_stack")

    def destroy_stack(self, stack, *args, **kwargs):
        """Abstract method."""
        not_implemented("destroy_stack")

    def get_stack_status(self, stack, *args, **kwargs):
        """Abstract method."""
        not_implemented("get_stack_status")

    def get_outputs(self, stack_name, *args, **kwargs):
        """Abstract method."""
        not_implemented("get_outputs")

    def get_output(self, stack, output):
        """Abstract method."""
        return self.get_outputs(stack)[output]


class Template(object):
    """CloudFormation stack template, which could be optionally uploaded to s3.

    Presence of the url attribute indicates that the template was uploaded to
    S3, and the uploaded template should be used for
    ``CreateStack``/``UpdateStack`` calls.

    """

    def __init__(self, url=None, body=None):
        """Instantiate class."""
        self.url = url
        self.body = body



"""Utility functions."""
from __future__ import print_function
from typing import Any, Dict, Iterator, List, Optional, Union  # noqa pylint: disable=unused-import

from contextlib import contextmanager
import hashlib
import importlib
import json
import os
import platform
import re
import stat
from subprocess import check_call
import sys
import six

AWS_ENV_VARS = ('AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY',
                'AWS_SESSION_TOKEN')
EMBEDDED_LIB_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'embedded'
)


class cached_property(object):  # noqa pylint: disable=invalid-name,too-few-public-methods
    """Decorator for creating cached properties.

    A property that is only computed once per instance and then replaces itself
    with an ordinary attribute. Deleting the attribute resets the property.
    Source: https://github.com/bottlepy/bottle/commit/fa7733e075da0d790d809aa3d2f53071897e6f76

    """

    def __init__(self, func):
        """Initialize class.

        Args:
            func (Callable): Method being decorated.

        """
        self.func = func

    def __get__(self, obj, _):
        """Attempt to get a cached value.

        Args:
            obj (Any): Instance of a class.

        Returns:
            Any

        """
        if obj is None:
            return self

        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value


# python2 supported pylint is unable to load six.moves correctly
class MutableMap(six.moves.collections_abc.MutableMapping):  # pylint: disable=no-member
    """Base class for mutable map objects."""

    def __init__(self, **kwargs):
        # type: (Dict[str, Any]) -> None
        """Initialize class.

        Provided ``kwargs`` are added to the object as attributes.

        Example:
            .. codeblock: python

                obj = MutableMap(**{'key': 'value'})
                print(obj.__dict__)
                # {'key': 'value'}

        """
        for key, value in kwargs.items():
            if isinstance(value, dict):
                setattr(self, key, MutableMap(**value))
            else:
                setattr(self, key, value)
        if kwargs:
            self._found_queries = MutableMap()

    @property
    def data(self):
        # type: () -> Dict[str, Any]
        """Sanitized output of __dict__.

        Removes anything that starts with ``_``.

        """
        result = {}
        for key, val in self.__dict__.items():
            if key.startswith('_'):
                continue
            if isinstance(val, MutableMap):
                result[key] = val.data
            else:
                result[key] = val
        return result

    def clear_found_cache(self):
        # type: () -> None
        """Clear _found_cache."""
        for _, val in self.__dict__.items():
            if isinstance(val, MutableMap):
                val.clear_found_cache()
        if hasattr(self, '_found_queries'):
            self._found_queries.clear()

    def find(self, query, default=None, ignore_cache=False):
        # type: (str, Any, bool) -> Any
        """Find a value in the map.

        Previously found queries are cached to increase search speed. The
        cached value should only be used if values are not expected to change.

        Args:
            query: A period delimited string that is split to search for
                nested values
            default: The value to return if the query was unsuccessful.
            ignore_cache: Ignore cached value.

        """
        if not hasattr(self, '_found_queries'):
            # if not created from kwargs, this attr won't exist yet
            # this is done to prevent endless recursion
            self._found_queries = MutableMap()

        if not ignore_cache:
            cached_result = self._found_queries.get(query, None)
            if cached_result:
                return cached_result

        split_query = query.split('.')

        if len(split_query) == 1:
            result = self.get(split_query[0], default)
            if result != default:
                self._found_queries[split_query[0]] = result
            return result

        nested_value = self.get(split_query[0])

        if not nested_value:
            return default

        nested_value = nested_value.find(split_query[1])

        try:
            nested_value = self[split_query[0]].find('.'.join(split_query[1:]),
                                                     default, ignore_cache)
            if nested_value != default:
                self._found_queries[query] = nested_value
            return nested_value
        except (AttributeError, KeyError):
            return default

    def get(self, key, default=None):
        # type: (str, Any) -> Any
        """Implement evaluation of self.get.

        Args:
            key: Attribute name to return the value for.
            default: Value to return if attribute is not found.

        """
        return getattr(self, key, default)

    def __bool__(self):
        # type: () -> bool
        """Implement evaluation of instances as a bool."""
        if self.data:
            return True
        return False

    def __contains__(self, value):
        # type: () -> bool
        """Implement evaluation of 'in' conditional."""
        return value in self.data

    __nonzero__ = __bool__  # python2 compatability

    def __getitem__(self, key):
        # type: (str) -> Any
        """Implement evaluation of self[key].

        Args:
            key: Attribute name to return the value for.

        Returns:
            The value associated with the provided key/attribute name.

        Raises:
            Attribute: If attribute does not exist on this object.

        Example:
            .. codeblock: python

                obj = MutableMap(**{'key': 'value'})
                print(obj['key'])
                # value

        """
        return getattr(self, key)

    def __setitem__(self, key, value):
        # type: (str, Any) -> None
        """Implement assignment to self[key].

        Args:
            key: Attribute name to associate with a value.
            value: Value of a key/attribute.

        Example:
            .. codeblock: python

                obj = MutableMap()
                obj['key'] = 'value'
                print(obj['key'])
                # value

        """
        if isinstance(value, dict):
            setattr(self, key, MutableMap(**value))
        else:
            setattr(self, key, value)

    def __delitem__(self, key):
        # type: (str) -> None
        """Implement deletion of self[key].

        Args:
            key: Attribute name to remove from the object.

        Example:
            .. codeblock: python

                obj = MutableMap(**{'key': 'value'})
                del obj['key']
                print(obj.__dict__)
                # {}

        """
        delattr(self, key)

    def __len__(self):
        # type: () -> int
        """Implement the built-in function len().

        Example:
            .. codeblock: python

                obj = MutableMap(**{'key': 'value'})
                print(len(obj))
                # 1

        """
        return len(self.__dict__)

    def __iter__(self):
        # type: () -> Iterator[Any]
        """Return iterator object that can iterate over all attributes.

        Example:
            .. codeblock: python

                obj = MutableMap(**{'key': 'value'})
                for k, v in obj.items():
                    print(f'{key}: {value}')
                # key: value

        """
        return iter(self.__dict__)

    def __str__(self):
        # type: () -> str
        """Return string representation of the object."""
        return json.dumps(self.data)


@contextmanager
def change_dir(newdir):
    """Change directory.

    Adapted from http://stackoverflow.com/a/24176022
    """
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)


def ensure_file_is_executable(path):
    """Exit if file is not executable."""
    if platform.system() != 'Windows' and (
            not stat.S_IXUSR & os.stat(path)[stat.ST_MODE]):
        print("Error: File %s is not executable" % path)
        sys.exit(1)


def load_object_from_string(fqcn):
    """Convert "." delimited strings to a python object.

    Given a "." delimited string representing the full path to an object
    (function, class, variable) inside a module, return that object.

    Example::

        load_object_from_string("os.path.basename")
        load_object_from_string("logging.Logger")
        load_object_from_string("LocalClassName")

    """
    module_path = "__main__"
    object_name = fqcn
    if "." in fqcn:
        module_path, object_name = fqcn.rsplit(".", 1)
        importlib.import_module(module_path)
    return getattr(sys.modules[module_path], object_name)


def merge_dicts(dict1, dict2, deep_merge=True):
    """Merge dict2 into dict1."""
    if deep_merge:
        if isinstance(dict1, list) and isinstance(dict2, list):
            return dict1 + dict2

        if not isinstance(dict1, dict) or not isinstance(dict2, dict):
            return dict2

        for key in dict2:
            dict1[key] = merge_dicts(dict1[key], dict2[key]) if key in dict1 else dict2[key]  # noqa pylint: disable=line-too-long
        return dict1
    dict3 = dict1.copy()
    dict3.update(dict2)
    return dict3
    # Alternate py3 version:
    # (tbd if it does or doesn't deep merge, and if that is needed)
    # if sys.version_info > (3, 4):
    #     return {**dict1, **dict2}


def extract_boto_args_from_env(env_vars):
    """Return boto3 client args dict with environment creds."""
    boto_args = {}
    for i in ['aws_access_key_id', 'aws_secret_access_key',
              'aws_session_token']:
        if env_vars.get(i.upper()):
            boto_args[i] = env_vars[i.upper()]
    return boto_args


def flatten_path_lists(env_dict, env_root=None):
    """Join paths in environment dict down to strings."""
    for (key, val) in env_dict.items():
        # Lists are presumed to be path components and will be turned back
        # to strings
        if isinstance(val, list):
            env_dict[key] = os.path.join(env_root, os.path.join(*val)) if (env_root and not os.path.isabs(os.path.join(*val))) else os.path.join(*val)  # noqa pylint: disable=line-too-long
    return env_dict


def merge_nested_environment_dicts(env_dicts, env_name=None, env_root=None):
    """Return single-level dictionary from dictionary of dictionaries."""
    # If the provided dictionary is just a single "level" (no nested
    # environments), it applies to all environments
    if all(isinstance(val, (six.string_types, list))
           for (_key, val) in env_dicts.items()):
        return flatten_path_lists(env_dicts, env_root)

    if env_name is None:
        if env_dicts.get('*'):
            return flatten_path_lists(env_dicts.get('*'), env_root)
        return {}

    if not env_dicts.get('*') and not env_dicts.get(env_name):
        return {}

    combined_dicts = merge_dicts(env_dicts.get('*', {}),
                                 env_dicts.get(env_name, {}))
    return flatten_path_lists(combined_dicts, env_root)


def find_cfn_output(key, outputs):
    """Return CFN output value."""
    for i in outputs:
        if i['OutputKey'] == key:
            return i['OutputValue']
    return None


def get_embedded_lib_path():
    """Return path of embedded libraries."""
    return os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'embedded'
    )


def get_hash_for_filename(filename, hashfile_path):
    """Return hash for filename in the hashfile."""
    filehash = ''
    with open(hashfile_path, 'r') as stream:
        for _cnt, line in enumerate(stream):
            if line.rstrip().endswith(filename):
                filehash = re.match(r'^[A-Za-z0-9]*', line).group(0)
                break
    if filehash:
        return filehash
    raise AttributeError("Filename %s not found in hash file" % filename)


@contextmanager
def ignore_exit_code_0():
    """Capture exit calls and ignore those with exit code 0."""
    try:
        yield
    except SystemExit as exit_exc:
        if exit_exc.code != 0:
            raise


def fix_windows_command_list(commands):
    # type: (List[str]) -> List[str]
    """Return command list with working Windows commands.

    npm on windows is npm.cmd, which will blow up
    subprocess.check_call(['npm', '...'])

    Similar issues arise when calling python apps like pipenv that will have
    a windows-only suffix applied to them
    """
    fully_qualified_cmd_path = which(commands[0])
    if fully_qualified_cmd_path:
        commands[0] = os.path.basename(fully_qualified_cmd_path)
    return commands


def run_commands(commands,  # type: List[Union[str, List[str], Dict[str, Union[str, List[str]]]]]
                 directory,  # type: str
                 env=None  # type: Optional[Dict[str, Union[str, int]]]
                ):  # noqa
    # type: (...) -> None
    """Run list of commands."""
    if env is None:
        env = os.environ.copy()
    for step in commands:
        if isinstance(step, (list, six.string_types)):
            execution_dir = directory
            raw_command = step
        elif step.get('command'):  # dictionary
            execution_dir = os.path.join(directory,
                                         step.get('cwd')) if step.get('cwd') else directory  # noqa pylint: disable=line-too-long
            raw_command = step['command']
        else:
            raise AttributeError("Invalid command step: %s" % step)
        command_list = raw_command.split(' ') if isinstance(raw_command, six.string_types) else raw_command  # noqa pylint: disable=line-too-long
        if platform.system().lower() == 'windows':
            command_list = fix_windows_command_list(command_list)

        with change_dir(execution_dir):
            failed_to_find_error = "Attempted to run \"%s\" and failed to find it (are you sure it is installed and added to your PATH?)" % command_list[0]  # noqa pylint: disable=line-too-long
            if sys.version_info[0] < 3:
                # Legacy exception version for python 2
                try:
                    check_call(command_list, env=env)
                except OSError:
                    print(failed_to_find_error, file=sys.stderr)
                    sys.exit(1)
            else:
                try:
                    check_call(command_list, env=env)
                # The noqa/pylint overrides can be dropped alongside python 2
                except FileNotFoundError:  # noqa pylint: disable=undefined-variable
                    print(failed_to_find_error, file=sys.stderr)
                    sys.exit(1)


def md5sum(filename):
    """Return MD5 hash of file."""
    md5 = hashlib.md5()
    with open(filename, 'rb') as stream:
        while True:
            data = stream.read(65536)  # 64kb chunks
            if not data:
                break
            md5.update(data)
    return md5.hexdigest()


def sha256sum(filename):
    """Return SHA256 hash of file."""
    sha256 = hashlib.sha256()
    mem_view = memoryview(bytearray(128 * 1024))
    with open(filename, 'rb', buffering=0) as stream:
        for i in iter(lambda: stream.readinto(mem_view), 0):
            sha256.update(mem_view[:i])
    return sha256.hexdigest()


def strip_leading_option_delim(args):
    """Remove leading -- if present.

    Using the "--" end of options syntax bypasses docopt's parsing of options.
    """
    if len(args) > 1:
        if args[0] == '--':
            return args[1:]
    return args


@contextmanager
def use_embedded_pkgs(embedded_lib_path=None):
    """Temporarily prepend embedded packages to sys.path."""
    if embedded_lib_path is None:
        embedded_lib_path = get_embedded_lib_path()

    old_sys_path = list(sys.path)
    sys.path.insert(
        1,  # https://stackoverflow.com/a/10097543
        embedded_lib_path
    )
    try:
        yield
    finally:
        sys.path = old_sys_path


def which(program):
    """Mimic 'which' command behavior."""
    def is_exe(fpath):
        """Determine if program exists and is executable."""
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    def get_extensions():
        """Get PATHEXT if the exist, otherwise use default."""
        exts = ['.COM;.EXE;.BAT;.CMD;.VBS;.VBE;.JS;.JSE;.WSF;.WSH;.MSC']

        if os.environ.get('PATHEXT', []):
            exts = os.environ['PATHEXT']

        return exts.split(';')

    fname, file_ext = os.path.splitext(program)
    fpath, fname = os.path.split(program)

    if not file_ext and platform.system().lower() == 'windows':
        fnames = [fname + ext for ext in get_extensions()]
    else:
        fnames = [fname]

    for i in fnames:
        if fpath:
            exe_file = os.path.join(fpath, i)
            if is_exe(exe_file):
                return exe_file
        else:
            for path in os.environ.get('PATH').split(os.pathsep) if 'PATH' in os.environ else [os.getcwd()]:  # noqa pylint: disable=line-too-long
                exe_file = os.path.join(path, i)
                if is_exe(exe_file):
                    return exe_file

    return None



"""Test deploying stacker."""
import os
from subprocess import check_output
import boto3

from integration_tests.test_commands.test_commands import Commands

KEY = "/runway/integration-test/stacker"
VALUE = "foo"


class TestRunStacker(Commands):
    """Tests run-stacker subcommand."""

    TEST_NAME = __name__

    def get_stack_path(self):
        """Gets the stacker itest path."""
        return os.path.join(
            os.path.dirname(os.path.dirname(__file__)),
            'fixtures',
            'stacker')

    def init(self):
        """Initialize test."""
        pass  # pylint: disable=unnecessary-pass

    def run(self):
        """Run tests."""
        path = self.get_stack_path()
        check_output(
            ['runway',
             'run-stacker',
             '--',
             'build',
             'stack.yaml'],
            cwd=path).decode()
        client = boto3.client('ssm',
                              region_name=self.environment['AWS_DEFAULT_REGION'])
        parameter = client.get_parameter(Name=KEY)
        assert parameter['Parameter']['Value'] == VALUE

    def teardown(self):
        """Teardown any created resources."""
        path = self.get_stack_path()
        check_output(
            ['runway',
             'run-stacker',
             '--',
             'destroy',
             'stack.yaml',
             '--force'],
            cwd=path).decode()



"""Tests for runway.cfngin.providers.aws.default."""
# pylint: disable=too-many-lines
import copy
import os.path
import random
import string
import sys
import threading
import unittest
from datetime import datetime

import boto3
from botocore.exceptions import ClientError, UnStubbedResponseError
from botocore.stub import Stubber
from mock import MagicMock, patch

from runway.cfngin import exceptions
from runway.cfngin.actions.diff import DictValue
from runway.cfngin.providers.aws import default
from runway.cfngin.providers.aws.default import (DEFAULT_CAPABILITIES,
                                                 MAX_TAIL_RETRIES, Provider,
                                                 ask_for_approval,
                                                 create_change_set,
                                                 generate_cloudformation_args,
                                                 output_full_changeset,
                                                 requires_replacement,
                                                 summarize_params_diff,
                                                 wait_till_change_set_complete)
from runway.cfngin.providers.base import Template
from runway.cfngin.session_cache import get_session
from runway.cfngin.stack import Stack
from runway.util import MutableMap

if sys.version_info.major < 3:
    from pathlib2 import Path  # pylint: disable=E
else:
    from pathlib import Path  # pylint: disable=E


def random_string(length=12):
    """Return a random string of variable length.

    Args:
        length (int): The # of characters to use in the random string.

    Returns:
        str: The random string.

    """
    return ''.join(
        [random.choice(string.ascii_letters) for _ in range(length)])


def generate_describe_stacks_stack(stack_name,
                                   creation_time=None,
                                   stack_status="CREATE_COMPLETE",
                                   tags=None,
                                   termination_protection=False):
    """Generate describe stacks stack."""
    tags = tags or []
    return {
        "StackName": stack_name,
        "StackId": stack_name,
        "CreationTime": creation_time or datetime(2015, 1, 1),
        "StackStatus": stack_status,
        "Tags": tags,
        "EnableTerminationProtection": termination_protection
    }


def generate_get_template(file_name='cfn_template.json',
                          stages_available=None):
    """Generate get template."""
    fixture_dir = os.path.join(os.path.dirname(__file__), '../../fixtures')
    with open(os.path.join(fixture_dir, file_name), 'r') as _file:
        return {
            "StagesAvailable": stages_available or ['Original'],
            "TemplateBody": _file.read()
        }


def generate_stack_object(stack_name, outputs=None):
    """Generate stack object."""
    mock_stack = MagicMock(['name', 'fqn', 'blueprint'])
    if not outputs:
        outputs = {
            "FakeOutput": {
                "Value": {"Ref": "FakeResource"}
            }
        }
    mock_stack.name = stack_name
    mock_stack.fqn = stack_name
    mock_stack.blueprint = MagicMock(['get_output_definitions'])
    mock_stack.blueprint.get_output_definitions = MagicMock(
        return_value=outputs
    )
    return mock_stack


def generate_resource_change(replacement=True):
    """Generate resource change."""
    resource_change = {
        "Action": "Modify",
        "Details": [],
        "LogicalResourceId": "Fake",
        "PhysicalResourceId": "arn:aws:fake",
        "Replacement": "True" if replacement else "False",
        "ResourceType": "AWS::Fake",
        "Scope": ["Properties"],
    }
    return {
        "ResourceChange": resource_change,
        "Type": "Resource",
    }


def generate_change_set_response(status, execution_status="AVAILABLE",
                                 changes=None, status_reason="FAKE"):
    """Generate change set response."""
    return {
        "ChangeSetName": "string",
        "ChangeSetId": "string",
        "StackId": "string",
        "StackName": "string",
        "Description": "string",
        "Parameters": [
            {
                "ParameterKey": "string",
                "ParameterValue": "string",
                "UsePreviousValue": False
            },
        ],
        "CreationTime": datetime(2015, 1, 1),
        "ExecutionStatus": execution_status,
        "Status": status,
        "StatusReason": status_reason,
        "NotificationARNs": [
            "string",
        ],
        "Capabilities": [
            "CAPABILITY_NAMED_IAM",
            "CAPABILITY_AUTO_EXPAND"
        ],
        "Tags": [
            {
                "Key": "string",
                "Value": "string"
            },
        ],
        "Changes": changes or [],
        "NextToken": "string"
    }


def generate_change(action="Modify", resource_type="EC2::Instance",
                    replacement="False", requires_recreation="Never"):
    """Generate a minimal change for a changeset."""
    return {
        "Type": "Resource",
        "ResourceChange": {
            "Action": action,
            "LogicalResourceId": random_string(),
            "PhysicalResourceId": random_string(),
            "ResourceType": resource_type,
            "Replacement": replacement,
            "Scope": ["Properties"],
            "Details": [
                {
                    "Target": {
                        "Attribute": "Properties",
                        "Name": random_string(),
                        "RequiresRecreation": requires_recreation
                    },
                    "Evaluation": "Static",
                    "ChangeSource": "ResourceReference",
                    "CausingEntity": random_string()
                },
            ]
        }
    }


class TestMethods(unittest.TestCase):
    """Tests for runway.cfngin.providers.aws.default."""

    def setUp(self):
        """Run before tests."""
        self.cfn = boto3.client("cloudformation")
        self.stubber = Stubber(self.cfn)

    def test_requires_replacement(self):
        """Test requires replacement."""
        changeset = [
            generate_resource_change(),
            generate_resource_change(replacement=False),
            generate_resource_change(),
        ]
        replacement = requires_replacement(changeset)
        self.assertEqual(len(replacement), 2)
        for resource in replacement:
            self.assertEqual(resource["ResourceChange"]["Replacement"], "True")

    def test_summarize_params_diff(self):
        """Test summarize params diff."""
        unmodified_param = DictValue("ParamA", "new-param-value",
                                     "new-param-value")
        modified_param = DictValue("ParamB", "param-b-old-value",
                                   "param-b-new-value-delta")
        added_param = DictValue("ParamC", None, "param-c-new-value")
        removed_param = DictValue("ParamD", "param-d-old-value", None)

        params_diff = [
            unmodified_param,
            modified_param,
            added_param,
            removed_param,
        ]
        self.assertEqual(summarize_params_diff([]), "")
        self.assertEqual(summarize_params_diff(params_diff), '\n'.join([
            "Parameters Added: ParamC",
            "Parameters Removed: ParamD",
            "Parameters Modified: ParamB\n",
        ]))

        only_modified_params_diff = [modified_param]
        self.assertEqual(summarize_params_diff(only_modified_params_diff),
                         "Parameters Modified: ParamB\n")

        only_added_params_diff = [added_param]
        self.assertEqual(summarize_params_diff(only_added_params_diff),
                         "Parameters Added: ParamC\n")

        only_removed_params_diff = [removed_param]
        self.assertEqual(summarize_params_diff(only_removed_params_diff),
                         "Parameters Removed: ParamD\n")

    def test_ask_for_approval(self):
        """Test ask for approval."""
        get_input_path = "runway.cfngin.ui.get_raw_input"
        with patch(get_input_path, return_value="y"):
            self.assertIsNone(ask_for_approval([], [], None))

        for v in ("n", "N", "x", "\n"):
            with patch(get_input_path, return_value=v):
                with self.assertRaises(exceptions.CancelExecution):
                    ask_for_approval([], [])

        with patch(get_input_path, side_effect=["v", "n"]) as mock_get_input:
            with patch(
                    'runway.cfngin.providers.aws.default.output_full_changeset'
            ) as mock_full_changeset:
                with self.assertRaises(exceptions.CancelExecution):
                    ask_for_approval([], [], True)
                self.assertEqual(mock_full_changeset.call_count, 1)
            self.assertEqual(mock_get_input.call_count, 2)

    def test_ask_for_approval_with_params_diff(self):
        """Test ask for approval with params diff."""
        get_input_path = "runway.cfngin.ui.get_raw_input"
        params_diff = [
            DictValue('ParamA', None, 'new-param-value'),
            DictValue('ParamB', 'param-b-old-value', 'param-b-new-value-delta')
        ]
        with patch(get_input_path, return_value="y"):
            self.assertIsNone(ask_for_approval([], params_diff, None))

        for v in ("n", "N", "x", "\n"):
            with patch(get_input_path, return_value=v):
                with self.assertRaises(exceptions.CancelExecution):
                    ask_for_approval([], params_diff)

        with patch(get_input_path, side_effect=["v", "n"]) as mock_get_input:
            with patch(
                    'runway.cfngin.providers.aws.default.output_full_changeset'
            ) as mock_full_changeset:
                with self.assertRaises(exceptions.CancelExecution):
                    ask_for_approval([], params_diff, True)
                self.assertEqual(mock_full_changeset.call_count, 1)
            self.assertEqual(mock_get_input.call_count, 2)

    @patch("runway.cfngin.providers.aws.default.format_params_diff")
    @patch('runway.cfngin.providers.aws.default.yaml.safe_dump')
    def test_output_full_changeset(self, mock_safe_dump, patched_format):
        """Test output full changeset."""
        get_input_path = "runway.cfngin.ui.get_raw_input"

        safe_dump_counter = 0

        for v in ['y', 'v', 'Y', 'V']:
            with patch(get_input_path, return_value=v) as prompt:
                self.assertIsNone(output_full_changeset(full_changeset=[],
                                                        params_diff=[],
                                                        fqn=None))
                self.assertEqual(prompt.call_count, 1)
                safe_dump_counter += 1
                self.assertEqual(mock_safe_dump.call_count, safe_dump_counter)
                self.assertEqual(patched_format.call_count, 0)

        for v in ['n', 'N']:
            with patch(get_input_path, return_value=v) as prompt:
                output_full_changeset(full_changeset=[], params_diff=[],
                                      answer=None, fqn=None)
                self.assertEqual(prompt.call_count, 1)
                self.assertEqual(mock_safe_dump.call_count, safe_dump_counter)
                self.assertEqual(patched_format.call_count, 0)

        with self.assertRaises(exceptions.CancelExecution):
            output_full_changeset(full_changeset=[], params_diff=[],
                                  answer='x', fqn=None)

        output_full_changeset(full_changeset=[], params_diff=['mock'],
                              answer='y', fqn=None)
        safe_dump_counter += 1
        self.assertEqual(mock_safe_dump.call_count, safe_dump_counter)
        self.assertEqual(patched_format.call_count, 1)

    def test_wait_till_change_set_complete_success(self):
        """Test wait till change set complete success."""
        self.stubber.add_response(
            "describe_change_set",
            generate_change_set_response("CREATE_COMPLETE")
        )
        with self.stubber:
            wait_till_change_set_complete(self.cfn, "FAKEID")

        self.stubber.add_response(
            "describe_change_set",
            generate_change_set_response("FAILED")
        )
        with self.stubber:
            wait_till_change_set_complete(self.cfn, "FAKEID")

    def test_wait_till_change_set_complete_failed(self):
        """Test wait till change set complete failed."""
        # Need 2 responses for try_count
        for _ in range(2):
            self.stubber.add_response(
                "describe_change_set",
                generate_change_set_response("CREATE_PENDING")
            )
        with self.stubber:
            with self.assertRaises(exceptions.ChangesetDidNotStabilize):
                wait_till_change_set_complete(self.cfn, "FAKEID", try_count=2,
                                              sleep_time=.1)

    def test_create_change_set_stack_did_not_change(self):
        """Test create change set stack did not change."""
        self.stubber.add_response(
            "create_change_set",
            {'Id': 'CHANGESETID', 'StackId': 'STACKID'}
        )

        self.stubber.add_response(
            "describe_change_set",
            generate_change_set_response(
                "FAILED", status_reason="Stack didn't contain changes."
            )
        )

        self.stubber.add_response(
            "delete_change_set",
            {},
            expected_params={"ChangeSetName": "CHANGESETID"}
        )

        with self.stubber:
            with self.assertRaises(exceptions.StackDidNotChange):
                create_change_set(
                    cfn_client=self.cfn, fqn="my-fake-stack",
                    template=Template(url="http://fake.template.url.com/"),
                    parameters=[], tags=[]
                )

    def test_create_change_set_unhandled_failed_status(self):
        """Test create change set unhandled failed status."""
        self.stubber.add_response(
            "create_change_set",
            {'Id': 'CHANGESETID', 'StackId': 'STACKID'}
        )

        self.stubber.add_response(
            "describe_change_set",
            generate_change_set_response(
                "FAILED", status_reason="Some random bad thing."
            )
        )

        with self.stubber:
            with self.assertRaises(exceptions.UnhandledChangeSetStatus):
                create_change_set(
                    cfn_client=self.cfn, fqn="my-fake-stack",
                    template=Template(url="http://fake.template.url.com/"),
                    parameters=[], tags=[]
                )

    def test_create_change_set_bad_execution_status(self):
        """Test create change set bad execution status."""
        self.stubber.add_response(
            "create_change_set",
            {'Id': 'CHANGESETID', 'StackId': 'STACKID'}
        )

        self.stubber.add_response(
            "describe_change_set",
            generate_change_set_response(
                status="CREATE_COMPLETE", execution_status="UNAVAILABLE",
            )
        )

        with self.stubber:
            with self.assertRaises(exceptions.UnableToExecuteChangeSet):
                create_change_set(
                    cfn_client=self.cfn, fqn="my-fake-stack",
                    template=Template(url="http://fake.template.url.com/"),
                    parameters=[], tags=[]
                )

    def test_generate_cloudformation_args(self):
        """Test generate cloudformation args."""
        stack_name = "mystack"
        template_url = "http://fake.s3url.com/blah.json"
        template_body = '{"fake_body": "woot"}'
        std_args = {
            "stack_name": stack_name, "parameters": [], "tags": [],
            "template": Template(url=template_url)}
        std_return = {"StackName": stack_name, "Parameters": [], "Tags": [],
                      "Capabilities": DEFAULT_CAPABILITIES,
                      "TemplateURL": template_url}
        result = generate_cloudformation_args(**std_args)
        self.assertEqual(result, std_return)

        result = generate_cloudformation_args(service_role="FakeRole",
                                              **std_args)
        service_role_result = copy.deepcopy(std_return)
        service_role_result["RoleARN"] = "FakeRole"
        self.assertEqual(result, service_role_result)

        result = generate_cloudformation_args(change_set_name="MyChanges",
                                              **std_args)
        change_set_result = copy.deepcopy(std_return)
        change_set_result["ChangeSetName"] = "MyChanges"
        self.assertEqual(result, change_set_result)

        # Check stack policy
        stack_policy = Template(body="{}")
        result = generate_cloudformation_args(stack_policy=stack_policy,
                                              **std_args)
        stack_policy_result = copy.deepcopy(std_return)
        stack_policy_result["StackPolicyBody"] = "{}"
        self.assertEqual(result, stack_policy_result)

        # If not TemplateURL is provided, use TemplateBody
        std_args["template"] = Template(body=template_body)
        template_body_result = copy.deepcopy(std_return)
        del template_body_result["TemplateURL"]
        template_body_result["TemplateBody"] = template_body
        result = generate_cloudformation_args(**std_args)
        self.assertEqual(result, template_body_result)


class TestProviderDefaultMode(unittest.TestCase):
    """Tests for runway.cfngin.providers.aws.default default mode."""

    def setUp(self):
        """Run before tests."""
        region = "us-east-1"
        self.session = get_session(region=region)
        self.provider = Provider(
            self.session, region=region, recreate_failed=False)
        self.stubber = Stubber(self.provider.cloudformation)

    def test_create_stack_no_changeset(self):
        """Test create_stack, no changeset, template url."""
        stack_name = 'fake_stack'
        template = Template(url="http://fake.template.url.com/")
        parameters = []
        tags = []

        expected_args = generate_cloudformation_args(stack_name,
                                                     parameters,
                                                     tags,
                                                     template)
        expected_args['EnableTerminationProtection'] = False

        self.stubber.add_response(
            'create_stack',
            {'StackId': stack_name},
            expected_args
        )

        with self.stubber:
            self.provider.create_stack(stack_name, template, parameters, tags)
        self.stubber.assert_no_pending_responses()

    @patch('runway.cfngin.providers.aws.default.Provider.update_termination_protection')
    @patch('runway.cfngin.providers.aws.default.create_change_set')
    def test_create_stack_with_changeset(self, patched_create_change_set,
                                         patched_update_term):
        """Test create_stack, force changeset, termination protection."""
        stack_name = 'fake_stack'
        template_path = Path('./tests/cfngin/fixtures/cfn_template.yaml')
        template = Template(body=template_path.read_text())
        parameters = []
        tags = []

        changeset_id = 'CHANGESETID'

        patched_create_change_set.return_value = ([], changeset_id)

        self.stubber.add_response(
            'execute_change_set',
            {},
            {'ChangeSetName': changeset_id}
        )

        with self.stubber:
            self.provider.create_stack(stack_name, template, parameters, tags,
                                       force_change_set=True,
                                       termination_protection=True)
        self.stubber.assert_no_pending_responses()

        patched_create_change_set.assert_called_once_with(
            self.provider.cloudformation, stack_name, template, parameters,
            tags, 'CREATE', service_role=self.provider.service_role
        )
        patched_update_term.assert_called_once_with(stack_name, True)

    def test_destroy_stack(self):
        """Test destroy stack."""
        stack = {'StackName': 'MockStack'}

        self.stubber.add_response('delete_stack', {}, stack)

        with self.stubber:
            self.assertIsNone(self.provider.destroy_stack(stack))
            self.stubber.assert_no_pending_responses()

    def test_get_stack_stack_does_not_exist(self):
        """Test get stack stack does not exist."""
        stack_name = "MockStack"
        self.stubber.add_client_error(
            "describe_stacks",
            service_error_code="ValidationError",
            service_message="Stack with id %s does not exist" % stack_name,
            expected_params={"StackName": stack_name}
        )

        with self.assertRaises(exceptions.StackDoesNotExist):
            with self.stubber:
                self.provider.get_stack(stack_name)

    def test_get_stack_stack_exists(self):
        """Test get stack stack exists."""
        stack_name = "MockStack"
        stack_response = {
            "Stacks": [generate_describe_stacks_stack(stack_name)]
        }
        self.stubber.add_response(
            "describe_stacks",
            stack_response,
            expected_params={"StackName": stack_name}
        )

        with self.stubber:
            response = self.provider.get_stack(stack_name)

        self.assertEqual(response["StackName"], stack_name)

    def test_select_destroy_method(self):
        """Test select destroy method."""
        for i in [[{'force_interactive': False},
                   self.provider.noninteractive_destroy_stack],
                  [{'force_interactive': True},
                   self.provider.interactive_destroy_stack]]:
            self.assertEqual(self.provider.select_destroy_method(**i[0]),
                             i[1])

    def test_select_update_method(self):
        """Test select update method."""
        for i in [[{'force_interactive': True,
                    'force_change_set': False},
                   self.provider.interactive_update_stack],
                  [{'force_interactive': False,
                    'force_change_set': False},
                   self.provider.default_update_stack],
                  [{'force_interactive': False,
                    'force_change_set': True},
                   self.provider.noninteractive_changeset_update],
                  [{'force_interactive': True,
                    'force_change_set': True},
                   self.provider.interactive_update_stack]]:
            self.assertEqual(
                self.provider.select_update_method(**i[0]),
                i[1]
            )

    def test_prepare_stack_for_update_completed(self):
        """Test prepare stack for update completed."""
        stack_name = "MockStack"
        stack = generate_describe_stacks_stack(
            stack_name, stack_status="UPDATE_COMPLETE")

        with self.stubber:
            self.assertTrue(
                self.provider.prepare_stack_for_update(stack, []))

    def test_prepare_stack_for_update_in_progress(self):
        """Test prepare stack for update in progress."""
        stack_name = "MockStack"
        stack = generate_describe_stacks_stack(
            stack_name, stack_status="UPDATE_IN_PROGRESS")

        with self.assertRaises(exceptions.StackUpdateBadStatus) as raised:
            with self.stubber:
                self.provider.prepare_stack_for_update(stack, [])

            self.assertIn('in-progress', str(raised.exception))

    def test_prepare_stack_for_update_non_recreatable(self):
        """Test prepare stack for update non recreatable."""
        stack_name = "MockStack"
        stack = generate_describe_stacks_stack(
            stack_name, stack_status="REVIEW_IN_PROGRESS")

        with self.assertRaises(exceptions.StackUpdateBadStatus) as raised:
            with self.stubber:
                self.provider.prepare_stack_for_update(stack, [])

        self.assertIn('Unsupported state', str(raised.exception))

    def test_prepare_stack_for_update_disallowed(self):
        """Test prepare stack for update disallowed."""
        stack_name = "MockStack"
        stack = generate_describe_stacks_stack(
            stack_name, stack_status="ROLLBACK_COMPLETE")

        with self.assertRaises(exceptions.StackUpdateBadStatus) as raised:
            with self.stubber:
                self.provider.prepare_stack_for_update(stack, [])

        self.assertIn('re-creation is disabled', str(raised.exception))
        # Ensure we point out to the user how to enable re-creation
        self.assertIn('--recreate-failed', str(raised.exception))

    def test_prepare_stack_for_update_bad_tags(self):
        """Test prepare stack for update bad tags."""
        stack_name = "MockStack"
        stack = generate_describe_stacks_stack(
            stack_name, stack_status="ROLLBACK_COMPLETE")

        self.provider.recreate_failed = True

        with self.assertRaises(exceptions.StackUpdateBadStatus) as raised:
            with self.stubber:
                self.provider.prepare_stack_for_update(
                    stack,
                    tags=[{'Key': 'cfngin_namespace', 'Value': 'test'}])

        self.assertIn('tags differ', str(raised.exception).lower())

    def test_prepare_stack_for_update_recreate(self):
        """Test prepare stack for update recreate."""
        stack_name = "MockStack"
        stack = generate_describe_stacks_stack(
            stack_name, stack_status="ROLLBACK_COMPLETE")

        self.stubber.add_response(
            "delete_stack",
            {},
            expected_params={"StackName": stack_name}
        )

        self.provider.recreate_failed = True

        with self.stubber:
            self.assertFalse(
                self.provider.prepare_stack_for_update(stack, []))

    def test_noninteractive_changeset_update_no_stack_policy(self):
        """Test noninteractive changeset update no stack policy."""
        stack_name = "MockStack"

        self.stubber.add_response(
            "create_change_set",
            {'Id': 'CHANGESETID', 'StackId': 'STACKID'}
        )
        changes = []
        changes.append(generate_change())

        self.stubber.add_response(
            "describe_change_set",
            generate_change_set_response(
                status="CREATE_COMPLETE", execution_status="AVAILABLE",
                changes=changes,
            )
        )

        self.stubber.add_response("execute_change_set", {})

        with self.stubber:
            self.provider.noninteractive_changeset_update(
                fqn=stack_name,
                template=Template(url="http://fake.template.url.com/"),
                old_parameters=[],
                parameters=[], stack_policy=None, tags=[],
            )

    def test_noninteractive_changeset_update_with_stack_policy(self):
        """Test noninteractive changeset update with stack policy."""
        stack_name = "MockStack"

        self.stubber.add_response(
            "create_change_set",
            {'Id': 'CHANGESETID', 'StackId': 'STACKID'}
        )
        changes = []
        changes.append(generate_change())

        self.stubber.add_response(
            "describe_change_set",
            generate_change_set_response(
                status="CREATE_COMPLETE", execution_status="AVAILABLE",
                changes=changes,
            )
        )

        self.stubber.add_response("set_stack_policy", {})

        self.stubber.add_response("execute_change_set", {})

        with self.stubber:
            self.provider.noninteractive_changeset_update(
                fqn=stack_name,
                template=Template(url="http://fake.template.url.com/"),
                old_parameters=[],
                parameters=[], stack_policy=Template(body="{}"), tags=[],
            )

    def test_noninteractive_destroy_stack_termination_protected(self):
        """Test noninteractive_destroy_stack with termination protection."""
        self.stubber.add_client_error('delete_stack')

        with self.stubber, self.assertRaises(ClientError):
            self.provider.noninteractive_destroy_stack('fake-stack')
        self.stubber.assert_no_pending_responses()

    @patch('runway.cfngin.providers.aws.default.output_full_changeset')
    def test_get_stack_changes_update(self, mock_output_full_cs):
        """Test get stack changes update."""
        stack_name = "MockStack"
        mock_stack = generate_stack_object(stack_name)

        self.stubber.add_response(
            'describe_stacks',
            {'Stacks': [generate_describe_stacks_stack(stack_name)]}
        )
        self.stubber.add_response(
            'get_template',
            generate_get_template('cfn_template.yaml')
        )
        self.stubber.add_response(
            "create_change_set",
            {'Id': 'CHANGESETID', 'StackId': stack_name}
        )
        changes = []
        changes.append(generate_change())

        self.stubber.add_response(
            "describe_change_set",
            generate_change_set_response(
                status="CREATE_COMPLETE", execution_status="AVAILABLE",
                changes=changes,
            )
        )
        self.stubber.add_response("delete_change_set", {})
        self.stubber.add_response(
            'describe_stacks',
            {'Stacks': [generate_describe_stacks_stack(stack_name)]}
        )

        with self.stubber:
            result = self.provider.get_stack_changes(
                stack=mock_stack, template=Template(
                    url="http://fake.template.url.com/"
                ), parameters=[], tags=[])

        mock_output_full_cs.assert_called_with(full_changeset=changes,
                                               params_diff=[],
                                               fqn=stack_name,
                                               answer='y')
        expected_outputs = {
            'FakeOutput': '<inferred-change: MockStack.FakeOutput={}>'.format(
                str({"Ref": "FakeResource"})
            )
        }
        self.assertEqual(self.provider.get_outputs(stack_name),
                         expected_outputs)
        self.assertEqual(result, expected_outputs)

    @patch('runway.cfngin.providers.aws.default.output_full_changeset')
    def test_get_stack_changes_create(self, mock_output_full_cs):
        """Test get stack changes create."""
        stack_name = "MockStack"
        mock_stack = generate_stack_object(stack_name)

        self.stubber.add_response(
            'describe_stacks',
            {'Stacks': [generate_describe_stacks_stack(
                stack_name, stack_status='REVIEW_IN_PROGRESS'
            )]}
        )
        self.stubber.add_response(
            "create_change_set",
            {'Id': 'CHANGESETID', 'StackId': stack_name}
        )
        changes = []
        changes.append(generate_change())

        self.stubber.add_response(
            "describe_change_set",
            generate_change_set_response(
                status="CREATE_COMPLETE", execution_status="AVAILABLE",
                changes=changes,
            )
        )
        self.stubber.add_response("delete_change_set", {})
        self.stubber.add_response(
            'describe_stacks',
            {'Stacks': [generate_describe_stacks_stack(
                stack_name, stack_status='REVIEW_IN_PROGRESS'
            )]}
        )
        self.stubber.add_response(
            'describe_stacks',
            {'Stacks': [generate_describe_stacks_stack(
                stack_name, stack_status='REVIEW_IN_PROGRESS'
            )]}
        )

        self.stubber.add_response("delete_stack", {})

        with self.stubber:
            self.provider.get_stack_changes(
                stack=mock_stack, template=Template(
                    url="http://fake.template.url.com/"
                ), parameters=[], tags=[])

        mock_output_full_cs.assert_called_with(full_changeset=changes,
                                               params_diff=[],
                                               fqn=stack_name,
                                               answer='y')

    def test_tail_stack_retry_on_missing_stack(self):
        """Test tail stack retry on missing stack."""
        stack_name = "SlowToCreateStack"
        stack = MagicMock(spec=Stack)
        stack.fqn = "my-namespace-{}".format(stack_name)

        default.TAIL_RETRY_SLEEP = .01

        # Ensure the stack never appears before we run out of retries
        for i in range(MAX_TAIL_RETRIES + 5):
            self.stubber.add_client_error(
                "describe_stack_events",
                service_error_code="ValidationError",
                service_message="Stack [{}] does not exist".format(stack_name),
                http_status_code=400,
                response_meta={"attempt": i + 1},
            )

        with self.stubber:
            try:
                self.provider.tail_stack(stack, threading.Event())
            except ClientError as exc:
                self.assertEqual(
                    exc.response["ResponseMetadata"]["attempt"],
                    MAX_TAIL_RETRIES
                )

    def test_tail_stack_retry_on_missing_stack_eventual_success(self):
        """Test tail stack retry on missing stack eventual success."""
        stack_name = "SlowToCreateStack"
        stack = MagicMock(spec=Stack)
        stack.fqn = "my-namespace-{}".format(stack_name)

        default.TAIL_RETRY_SLEEP = .01
        default.GET_EVENTS_SLEEP = .01

        received_events = []

        def mock_log_func(event):
            received_events.append(event)

        def valid_event_response(stack, event_id):
            return {
                "StackEvents": [
                    {
                        "StackId": stack.fqn + "12345",
                        "EventId": event_id,
                        "StackName": stack.fqn,
                        "Timestamp": datetime.now()
                    },
                ]
            }

        # Ensure the stack never appears before we run out of retries
        for i in range(3):
            self.stubber.add_client_error(
                "describe_stack_events",
                service_error_code="ValidationError",
                service_message="Stack [{}] does not exist".format(stack_name),
                http_status_code=400,
                response_meta={"attempt": i + 1},
            )

        self.stubber.add_response(
            "describe_stack_events",
            valid_event_response(stack, "InitialEvents")
        )

        self.stubber.add_response(
            "describe_stack_events",
            valid_event_response(stack, "Event1")
        )

        with self.stubber:
            try:
                self.provider.tail_stack(stack, threading.Event(),
                                         log_func=mock_log_func)
            except UnStubbedResponseError:
                # Eventually we run out of responses - could not happen in
                # regular execution
                # normally this would just be dealt with when the threads were
                # shutdown, but doing so here is a little difficult because
                # we can't control the `tail_stack` loop
                pass

        self.assertEqual(received_events[0]["EventId"], "Event1")

    def test_update_termination_protection(self):
        """Test update_termination_protection."""
        stack_name = 'fake-stack'
        test_cases = [
            MutableMap(aws=False, defined=True, expected=True),
            MutableMap(aws=True, defined=False, expected=False),
            MutableMap(aws=True, defined=True, expected=None),
            MutableMap(aws=False, defined=False, expected=None)
        ]

        for test in test_cases:
            self.stubber.add_response(
                'describe_stacks',
                {'Stacks': [
                    generate_describe_stacks_stack(
                        stack_name, termination_protection=test.aws
                    )
                ]},
                {'StackName': stack_name}
            )
            if isinstance(test.expected, bool):
                self.stubber.add_response(
                    'update_termination_protection',
                    {'StackId': stack_name},
                    {'EnableTerminationProtection': test.expected,
                     'StackName': stack_name}
                )
            with self.stubber:
                self.provider.update_termination_protection(stack_name,
                                                            test.defined)
            self.stubber.assert_no_pending_responses()


class TestProviderInteractiveMode(unittest.TestCase):
    """Tests for runway.cfngin.providers.aws.default interactive mode."""

    def setUp(self):
        """Run before tests."""
        region = "us-east-1"
        self.session = get_session(region=region)
        self.provider = Provider(
            self.session, interactive=True, recreate_failed=True)
        self.stubber = Stubber(self.provider.cloudformation)

    @patch('runway.cfngin.ui.get_raw_input')
    def test_interactive_destroy_stack(self, patched_input):
        """Test interactive_destroy_stack."""
        stack_name = 'fake-stack'
        stack = {'StackName': stack_name}
        patched_input.return_value = 'y'

        self.stubber.add_response('delete_stack', {}, stack)

        with self.stubber:
            self.assertIsNone(self.provider.interactive_destroy_stack(stack_name))
            self.stubber.assert_no_pending_responses()

    @patch('runway.cfngin.providers.aws.default.Provider.update_termination_protection')
    @patch('runway.cfngin.ui.get_raw_input')
    def test_interactive_destroy_stack_termination_protected(self,
                                                             patched_input,
                                                             patched_update_term):
        """Test interactive_destroy_stack with termination protection."""
        stack_name = 'fake-stack'
        stack = {'StackName': stack_name}
        patched_input.return_value = 'y'

        self.stubber.add_client_error('delete_stack',
                                      service_message='TerminationProtection')
        self.stubber.add_response('delete_stack', {}, stack)

        with self.stubber:
            self.provider.interactive_destroy_stack(stack_name, approval='y')
        self.stubber.assert_no_pending_responses()
        patched_input.assert_called_once()
        patched_update_term.assert_called_once_with(stack_name, False)

    @patch('runway.cfngin.ui.get_raw_input')
    def test_destroy_stack_canceled(self, patched_input):
        """Test destroy stack canceled."""
        stack = {'StackName': 'MockStack'}
        patched_input.return_value = 'n'

        with self.assertRaises(exceptions.CancelExecution):
            self.provider.destroy_stack(stack)

    def test_successful_init(self):
        """Test successful init."""
        replacements = True
        provider = Provider(self.session, interactive=True,
                            replacements_only=replacements)
        self.assertEqual(provider.replacements_only, replacements)

    @patch('runway.cfngin.providers.aws.default.Provider.update_termination_protection')
    @patch("runway.cfngin.providers.aws.default.ask_for_approval")
    def test_update_stack_execute_success_no_stack_policy(self,
                                                          patched_approval,
                                                          patched_update_term):
        """Test update stack execute success no stack policy."""
        stack_name = "my-fake-stack"

        self.stubber.add_response(
            "create_change_set",
            {'Id': 'CHANGESETID', 'StackId': 'STACKID'}
        )
        changes = [generate_change()]

        self.stubber.add_response(
            "describe_change_set",
            generate_change_set_response(
                status="CREATE_COMPLETE", execution_status="AVAILABLE",
                changes=changes,
            )
        )

        self.stubber.add_response("execute_change_set", {})

        with self.stubber:
            self.provider.update_stack(
                fqn=stack_name,
                template=Template(url="http://fake.template.url.com/"),
                old_parameters=[],
                parameters=[], tags=[]
            )

        patched_approval.assert_called_once_with(full_changeset=changes,
                                                 params_diff=[],
                                                 include_verbose=True,
                                                 fqn=stack_name)
        patched_update_term.assert_called_once_with(stack_name, False)

    @patch('runway.cfngin.providers.aws.default.Provider.update_termination_protection')
    @patch("runway.cfngin.providers.aws.default.ask_for_approval")
    def test_update_stack_execute_success_with_stack_policy(self,
                                                            patched_approval,
                                                            patched_update_term):
        """Test update stack execute success with stack policy."""
        stack_name = "my-fake-stack"

        self.stubber.add_response(
            "create_change_set",
            {'Id': 'CHANGESETID', 'StackId': 'STACKID'}
        )
        changes = [generate_change()]

        self.stubber.add_response(
            "describe_change_set",
            generate_change_set_response(
                status="CREATE_COMPLETE", execution_status="AVAILABLE",
                changes=changes,
            )
        )

        self.stubber.add_response("set_stack_policy", {})

        self.stubber.add_response("execute_change_set", {})

        with self.stubber:
            self.provider.update_stack(
                fqn=stack_name,
                template=Template(url="http://fake.template.url.com/"),
                old_parameters=[],
                parameters=[], tags=[],
                stack_policy=Template(body="{}"),
            )

        patched_approval.assert_called_once_with(full_changeset=changes,
                                                 params_diff=[],
                                                 include_verbose=True,
                                                 fqn=stack_name)
        patched_update_term.assert_called_once_with(stack_name, False)

    def test_select_destroy_method(self):
        """Test select destroy method."""
        for i in [[{'force_interactive': False},
                   self.provider.interactive_destroy_stack],
                  [{'force_interactive': True},
                   self.provider.interactive_destroy_stack]]:
            self.assertEqual(self.provider.select_destroy_method(**i[0]),
                             i[1])

    def test_select_update_method(self):
        """Test select update method."""
        for i in [[{'force_interactive': False,
                    'force_change_set': False},
                   self.provider.interactive_update_stack],
                  [{'force_interactive': True,
                    'force_change_set': False},
                   self.provider.interactive_update_stack],
                  [{'force_interactive': False,
                    'force_change_set': True},
                   self.provider.interactive_update_stack],
                  [{'force_interactive': True,
                    'force_change_set': True},
                   self.provider.interactive_update_stack]]:
            self.assertEqual(
                self.provider.select_update_method(**i[0]),
                i[1]
            )

    @patch('runway.cfngin.providers.aws.default.output_full_changeset')
    @patch('runway.cfngin.providers.aws.default.output_summary')
    def test_get_stack_changes_interactive(self, mock_output_summary,
                                           mock_output_full_cs):
        """Test get stack changes interactive."""
        stack_name = "MockStack"
        mock_stack = generate_stack_object(stack_name)

        self.stubber.add_response(
            'describe_stacks',
            {'Stacks': [generate_describe_stacks_stack(stack_name)]}
        )
        self.stubber.add_response(
            'get_template',
            generate_get_template('cfn_template.yaml')
        )
        self.stubber.add_response(
            "create_change_set",
            {'Id': 'CHANGESETID', 'StackId': stack_name}
        )
        changes = []
        changes.append(generate_change())

        self.stubber.add_response(
            "describe_change_set",
            generate_change_set_response(
                status="CREATE_COMPLETE", execution_status="AVAILABLE",
                changes=changes,
            )
        )
        self.stubber.add_response("delete_change_set", {})
        self.stubber.add_response(
            'describe_stacks',
            {'Stacks': [generate_describe_stacks_stack(stack_name)]}
        )

        with self.stubber:
            self.provider.get_stack_changes(
                stack=mock_stack, template=Template(
                    url="http://fake.template.url.com/"
                ), parameters=[], tags=[])

        mock_output_summary.assert_called_with(stack_name, 'changes',
                                               changes, [],
                                               replacements_only=False)
        mock_output_full_cs.assert_called_with(full_changeset=changes,
                                               params_diff=[],
                                               fqn=stack_name)



"""Test AWS Lambda hook."""
# flake8: noqa
# pylint: disable=invalid-name
from os import path
from os.path import basename

import boto3
from send2trash import send2trash

from integration_tests.test_cfngin.test_cfngin import Cfngin

FILE_BASENAME = '.'.join(basename(__file__).split('.')[:-1])


class TestAwsLambda(Cfngin):
    """Test AWS Lambda Hook from cfngin.

    Requires valid AWS Credentials.

    """

    REQUIRED_FIXTURE_FILES = [FILE_BASENAME + '.yaml']
    TEST_NAME = __name__

    def invoke_lambda(self, client, func_name):
        """Verify lambda function deployed successfully."""
        self.logger.info('Invoking lambda function: %s', func_name)
        resp = client.invoke(
            FunctionName=func_name,
            InvocationType='RequestResponse'
        )

        return resp['StatusCode']

    def _build(self):
        """Execute and assert initial build."""
        self.set_environment('dev')
        code, _stdout, _stderr = self.runway_cmd('deploy')
        assert code == 0, 'exit code should be zero'

    def run(self):
        """Run test."""
        self.copy_fixtures()
        self._build()

        client = boto3.client('lambda', region_name=self.region)

        functions = ['dockerizepip-integrationtest',
                     'nondockerizepip-integrationtest']
        for func_name in functions:
            assert self.invoke_lambda(client, func_name) == 200, \
                f'{self.TEST_NAME}: Execution of lambda {func_name} failed'

    def teardown(self):
        """Teardown test."""
        self.runway_cmd('destroy')
        venv_dir = path.join(self.fixture_dir,
                             'lambda_src/dockerize_src/.venv')
        if path.isdir(venv_dir):
            send2trash(venv_dir)
        self.cleanup_fixtures()



"""Add all configured headers (CloudFront compatable) to origin response."""
from shared import as_cloud_front_headers, get_config  # pylint: disable=import-error

CONFIG = get_config()


def handler(event, _context):
    """Handle adding the headers to the origin response.

    Keyword Args:
        event (Any): The Lambda Event
    """
    headers = CONFIG.get('http_headers')
    # Format to be CloudFront compatable
    configured_headers = as_cloud_front_headers(headers)
    response = event['Records'][0]['cf']['response']
    response['headers'].update(configured_headers)
    return response



"""Parse the given Cognito/Local authorization information.

Ensure the parsed Cognito and Local authorization information from the retrieved
query string parameter and cookie headers matches our expectations. If it doesn't
then inform the user of a bad request, otherwise retrieve the Cognito tokens to
add to the cookie headers.
"""

import json
import logging
import traceback
import base64
from urllib.parse import parse_qs  # noqa pylint: disable=no-name-in-module,import-error

from shared import (create_error_html, extract_and_parse_cookies, get_config,  # noqa pylint: disable=import-error
                    get_cookie_headers, http_post_with_retry)

LOGGER = logging.getLogger(__name__)
CONFIG = get_config()


def handler(event, _context):
    """Handle the authorization parsing.

    Keyword Args:
        event (Any): The Lambda Event
    """
    request = event['Records'][0]['cf']['request']
    domain_name = request['headers']['host'][0]['value']
    redirected_from_uri = 'https://%s' % domain_name

    # Attempt to parse the request and retrieve authorization
    # tokens to integrate with our header cookies
    try:
        qsp = parse_qs(request.get('querystring'))
        # Authorization code given by Cognito
        code = qsp.get('code')
        # The requested URI and current nonce information
        state = json.loads(
            base64.urlsafe_b64decode(qsp.get('state')[0]).decode()
        )

        # Missing required components
        if not code or not state:
            msg = 'Invalid query string. ' + \
                  'Your query string should include parameters "state" and "code"'
            LOGGER.info(msg)
            raise Exception(msg)

        current_nonce = state.get('nonce')
        requested_uri = state.get('requestedUri')
        redirected_from_uri = requested_uri or "/"
        # Get all the cookies from the headers
        cookies = extract_and_parse_cookies(
            request.get('headers'),
            CONFIG.get('client_id')
        )
        # Retrieve the original nonce value as well as our PKCE code
        original_nonce = cookies.get('nonce')
        pkce = cookies.get('pkce')

        # If we're missing one of the nonces, or they don't match, cause an error
        if not current_nonce or not original_nonce or current_nonce != original_nonce:
            # No original nonce? CSRF violation
            if not original_nonce:
                msg = "Your browser didn't send the nonce cookie along, " + \
                      "but it is required for security (prevent CSRF)"
                LOGGER.error(msg)
                raise Exception(msg)
            # Nonce's don't match
            msg = "Nonce Mismatch"
            LOGGER.error(msg)
            raise Exception(msg)

        payload = {
            'grant_type': 'authorization_code',
            'client_id': CONFIG.get('client_id'),
            'redirect_uri': 'https://%s%s' % (domain_name, CONFIG.get('redirect_path_sign_in')),
            'code': code[0],
            'code_verifier': pkce
        }

        # Request tokens from our Cognito Authorization Domain
        tokens = http_post_with_retry(
            ('https://%s/oauth2/token' % CONFIG.get('cognito_auth_domain')),
            payload,
            {"Content-Type": "application/x-www-form-urlencoded"}
        )

        if not tokens:
            raise Exception('Was not able to obtain tokens from Cognito')

        headers = {
            'location': [
                {
                    'key': 'location',
                    'value': redirected_from_uri
                }
            ],
            'set-cookie': get_cookie_headers(
                CONFIG.get("client_id"),
                CONFIG.get("oauth_scopes"),
                tokens,
                domain_name,
                CONFIG.get('cookie_settings')
            )
        }
        headers.update(CONFIG.get('cloud_front_headers'))
        # Redirect user to the originally requested uri with the
        # token header cookies
        response = {
            'status': '307',
            'statusDescription': 'Temporary Redirect',
            'headers': headers
        }
        return response
    except Exception as err:  # pylint: disable=broad-except
        LOGGER.error(err)
        LOGGER.error(traceback.print_exc())

        headers = CONFIG.get('cloud_front_headers')
        headers['content-type'] = [
            {
                'key': 'Content-Type',
                'value': 'text/html; charset=UTF-8'
            }
        ]

        # Inform user of bad request and reason
        return {
            'body': create_error_html("Bad Request", err, redirected_from_uri),
            'status': '400',
            'headers': headers
        }



"""Resources to tokenize userdata."""
import re

from troposphere import GetAtt, Ref

HELPERS = {
    "Ref": Ref,
    "Fn::GetAtt": GetAtt
}

SPLIT_STRING = "(" + "|".join([r"%s\([^)]+\)" % h for h in HELPERS]) + ")"
REPLACE_STRING = \
    r"(?P<helper>%s)\((?P<args>['\"]?[^)]+['\"]?)+\)" % '|'.join(HELPERS)

SPLIT_RE = re.compile(SPLIT_STRING)
REPLACE_RE = re.compile(REPLACE_STRING)


def cf_tokenize(raw_userdata):
    """Parse UserData for Cloudformation helper functions.

    http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/user-data.html

    http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference.html

    http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/quickref-cloudformation.html#scenario-userdata-base64

    It breaks apart the given string at each recognized function (see
    ``HELPERS`` global variable) and instantiates the helper function objects
    in place of those.

    Args:
        raw_userdata (str): Unparsed userdata data string.

    Returns:
        List[str]: A list of string parts that is useful when used with
        :func:`troposphere.Join` and :func:`troposphere.Base64` to produce
        userdata.

    Example:
        .. code-block: python

            Base64(Join('', cf_tokenize(userdata_string)))

    """
    result = []
    parts = SPLIT_RE.split(raw_userdata)
    for part in parts:
        cf_func = REPLACE_RE.search(part)
        if cf_func:
            args = [a.strip("'\" ") for a in cf_func.group("args").split(",")]
            result.append(HELPERS[cf_func.group("helper")](*args).data)
        else:
            result.append(part)
    return result



"""Fake Docker API client."""
# pylint: disable=attribute-defined-outside-init,protected-access
import copy

import docker

from . import fake_api

try:
    from unittest import mock
except ImportError:
    import mock


class CopyReturnMagicMock(mock.MagicMock):
    """A MagicMock which deep copies every return value."""

    def _mock_call(self, *args, **kwargs):
        ret = super(CopyReturnMagicMock, self)._mock_call(*args, **kwargs)
        if isinstance(ret, (dict, list)):
            ret = copy.deepcopy(ret)
        return ret


def make_fake_api_client(overrides=None):
    """Return non-complete fake APIClient.

    This returns most of the default cases correctly, but most arguments that
    change behavior will not work.

    """
    if overrides is None:
        overrides = {}
    api_client = docker.APIClient()
    mock_attrs = {
        'build.return_value': fake_api.FAKE_CONTAINER_ID,
        'commit.return_value': fake_api.post_fake_commit()[1],
        'containers.return_value': fake_api.get_fake_containers()[1],
        'create_container.return_value':
            fake_api.post_fake_create_container()[1],
        'create_host_config.side_effect': api_client.create_host_config,
        'create_network.return_value': fake_api.post_fake_network()[1],
        'exec_create.return_value': fake_api.post_fake_exec_create()[1],
        'exec_start.return_value': fake_api.post_fake_exec_start()[1],
        'images.return_value': fake_api.get_fake_images()[1],
        'inspect_container.return_value':
            fake_api.get_fake_inspect_container()[1],
        'inspect_image.return_value': fake_api.get_fake_inspect_image()[1],
        'inspect_network.return_value': fake_api.get_fake_network()[1],
        'logs.return_value': [b'hello world\n'],
        'networks.return_value': fake_api.get_fake_network_list()[1],
        'start.return_value': None,
        'wait.return_value': {'StatusCode': 0},
    }
    mock_attrs.update(overrides)
    mock_client = CopyReturnMagicMock(**mock_attrs)

    mock_client._version = docker.constants.DEFAULT_DOCKER_API_VERSION
    return mock_client


def make_fake_client(overrides=None):
    """Return a Client with a fake APIClient."""
    client = docker.DockerClient()
    client.api = make_fake_api_client(overrides)
    return client



"""Load dependencies."""
import awacs.awslambda
import awacs.dynamodb
import awacs.logs
import awacs.sts
from awacs.aws import Allow, Policy, Principal, Statement
from troposphere import (AccountId, Export, GetAtt, Join, Output, Partition,
                         Ref, Region, Sub, awslambda, iam)

from runway.cfngin.blueprints.base import Blueprint
from runway.cfngin.blueprints.variables.types import CFNString


class BlueprintClass(Blueprint):
    """Blueprint for creating lambda function."""

    VARIABLES = {
        "Code": {
            "type": awslambda.Code,
            "description": "The troposphere.awslambda.Code object "
                           "returned by the aws lambda hook.",
        },
        "AppName": {
            "type": CFNString,
            "description": "Name of app"
        },
        "Entrypoint": {
            "type": CFNString,
            "description": "Lambda function entrypoint.",
            "default": "index.handler"
        }
    }

    def create_resources(self):
        """Create the resources."""
        template = self.template
        variables = self.get_variables()
        app_name = variables['AppName'].ref

        lambda_iam_role = template.add_resource(
            iam.Role(
                'LambdaRole',
                RoleName=Join('-',
                              [app_name,
                               'lambda-role']),
                AssumeRolePolicyDocument=Policy(
                    Version='2012-10-17',
                    Statement=[
                        Statement(
                            Effect=Allow,
                            Action=[awacs.sts.AssumeRole],
                            Principal=Principal('Service',
                                                ['lambda.amazonaws.com'])
                        )
                    ]
                ),
                Path='/service-role/',
                Policies=[
                    iam.Policy(
                        PolicyName=Join('-',
                                        [app_name,
                                         'lambda-policy']),
                        PolicyDocument=Policy(
                            Version='2012-10-17',
                            Statement=[
                                Statement(
                                    Action=[
                                        awacs.logs.CreateLogGroup,
                                        awacs.logs.CreateLogStream,
                                        awacs.logs.PutLogEvents
                                    ],
                                    Effect=Allow,
                                    Resource=[
                                        Join('', [
                                            'arn:',
                                            Partition,
                                            ':logs:',
                                            Region,
                                            ':',
                                            AccountId,
                                            ':log-group:/aws/lambda/',
                                            app_name,
                                            '-*'])
                                    ],
                                    Sid='WriteLogs'
                                )
                            ]
                        )
                    )
                ]
            )
        )

        lambda_function = template.add_resource(
            awslambda.Function(
                'LambdaFunction',
                Code=variables['Code'],
                Handler=variables['Entrypoint'].ref,
                Role=GetAtt(lambda_iam_role, 'Arn'),
                Runtime='python3.6',
                Timeout=30,
                MemorySize=128,
                FunctionName=Join('-', [app_name,
                                        'integrationtest'])
            )
        )

        template.add_output(
            Output(
                lambda_iam_role.title,
                Description='Lambda Role',
                Export=Export(Sub(
                    '${AWS::StackName}-%s' % lambda_iam_role.title
                )),
                Value=Ref(lambda_iam_role)
            )
        )
        template.add_output(
            Output(
                lambda_function.title,
                Description='Lambda Function',
                Export=Export(Sub(
                    '${AWS::StackName}-%s' % lambda_function.title
                )),
                Value=GetAtt(lambda_function, 'Arn')
            )
        )
        template.add_output(
            Output(
                lambda_function.title + 'Name',
                Description='Lambda Function Name',
                Export=Export(Sub(
                    '${AWS::StackName}-%sName' % lambda_function.title
                )),
                Value=Ref(lambda_function)
            )
        )

    def create_template(self):
        self.template.add_version('2010-09-09')
        self.template.add_description('Test Lambda'
                                      ' - {0}'.format('1.0.0'))
        self.create_resources()



"""CFNgin statuses."""
import operator


class Status(object):
    """CFNgin status base class.

    Attributes:
        name (str): Name of the status.
        code (int): Status code.
        reason (Optional[str]): Reason for the status.

    """

    def __init__(self, name, code, reason=None):
        """Instantiate class.

        Args:
            name (str): Name of the status.
            code (int): Status code.
            reason (Optional[str]): Reason for the status.

        """
        self.name = name
        self.code = code
        self.reason = reason or getattr(self, "reason", None)

    def _comparison(self, operator_, other):
        """Compare self to another object.

        Args:
            operator_ (operator): Operator for the comparison.
            other (Any): The other object to compare to self.

        Returns:
            bool: Whether the comparison passes.

        Raises:
            NotImplemented: other does not have ``code`` attribute.

        """
        if hasattr(other, "code"):
            return operator_(self.code, other.code)
        return NotImplemented

    def __eq__(self, other):
        """Compare if self is equal to another object.

        Returns:
            bool: Whether the comparison passes.

        """
        return self._comparison(operator.eq, other)

    def __ne__(self, other):
        """Compare if self is not equal to another object.

        Returns:
            bool: Whether the comparison passes.

        """
        return self._comparison(operator.ne, other)

    def __lt__(self, other):
        """Compare if self is less than another object.

        Returns:
            bool: Whether the comparison passes.

        """
        return self._comparison(operator.lt, other)

    def __gt__(self, other):
        """Compare if self is greater than another object.

        Returns:
            bool: Whether the comparison passes.

        """
        return self._comparison(operator.gt, other)

    def __le__(self, other):
        """Compare if self is less than or equal to another object.

        Returns:
            bool: Whether the comparison passes.

        """
        return self._comparison(operator.le, other)

    def __ge__(self, other):
        """Compare if self is greater than equal to another object.

        Returns:
            bool: Whether the comparison passes.

        """
        return self._comparison(operator.ge, other)


class CompleteStatus(Status):  # pylint: disable=too-few-public-methods
    """Status name of 'complete' with code of '2'."""

    def __init__(self, reason=None):
        """Instantiate class.

        Args:
            reason (Optional[str]): Reason for the status.

        """
        super(CompleteStatus, self).__init__("complete", 2, reason)


class FailedStatus(Status):  # pylint: disable=too-few-public-methods
    """Status name of 'failed' with code of '4'."""

    def __init__(self, reason=None):
        """Instantiate class.

        Args:
            reason (Optional[str]): Reason for the status.

        """
        super(FailedStatus, self).__init__("failed", 4, reason)


class PendingStatus(Status):  # pylint: disable=too-few-public-methods
    """Status name of 'pending' with code of '0'."""

    def __init__(self, reason=None):
        """Instantiate class.

        Args:
            reason (Optional[str]): Reason for the status.

        """
        super(PendingStatus, self).__init__("pending", 0, reason)


class SkippedStatus(Status):  # pylint: disable=too-few-public-methods
    """Status name of 'skipped' with code of '3'."""

    def __init__(self, reason=None):
        """Instantiate class.

        Args:
            reason (Optional[str]): Reason for the status.

        """
        super(SkippedStatus, self).__init__("skipped", 3, reason)


class SubmittedStatus(Status):  # pylint: disable=too-few-public-methods
    """Status name of 'submitted' with code of '1'."""

    def __init__(self, reason=None):
        """Instantiate class.

        Args:
            reason (Optional[str]): Reason for the status.

        """
        super(SubmittedStatus, self).__init__("submitted", 1, reason)


class DidNotChangeStatus(SkippedStatus):  # pylint: disable=too-few-public-methods
    """Skipped status with a reason of 'nochange'."""

    reason = "nochange"


class NotSubmittedStatus(SkippedStatus):  # pylint: disable=too-few-public-methods
    """Skipped status with a reason of 'disabled'."""

    reason = "disabled"


class NotUpdatedStatus(SkippedStatus):  # pylint: disable=too-few-public-methods
    """Skipped status with a reason of 'locked'."""

    reason = "locked"


class StackDoesNotExist(SkippedStatus):  # pylint: disable=too-few-public-methods
    """Skipped status with a reason of 'does not exist in cloudformation'."""

    reason = "does not exist in cloudformation"


PENDING = PendingStatus()
WAITING = PendingStatus(reason="waiting")
SUBMITTED = SubmittedStatus()
COMPLETE = CompleteStatus()
SKIPPED = SkippedStatus()
FAILED = FailedStatus()
INTERRUPTED = FailedStatus(reason="interrupted")



"""Test runway.module.cloudformation."""
# pylint: disable=protected-access,no-self-use
import os

from mock import patch

from runway.context import Context
from runway.module.cloudformation import CloudFormation
from runway.util import MutableMap


class TestCloudFormation(object):
    """Test runway.module.cloudformation.CloudFormation."""

    @property
    def generic_options(self):
        """Return generic module options."""
        return {
            'environment': True,
            'parameters': MutableMap(**{
                'test_key': 'test-value'
            })
        }

    @staticmethod
    def get_context(name='test', region='us-east-1'):
        """Create a basic Runway context object."""
        return Context(env_name=name,
                       env_region=region,
                       env_root=os.getcwd())

    @patch('runway.cfngin.CFNgin.deploy')
    def test_deploy(self, mock_action, tmp_path):
        """Test deploy."""
        module = CloudFormation(self.get_context(),
                                str(tmp_path),
                                self.generic_options)
        module.deploy()
        mock_action.assert_called_once()

    @patch('runway.cfngin.CFNgin.destroy')
    def test_destroy(self, mock_action, tmp_path):
        """Test destroy."""
        module = CloudFormation(self.get_context(),
                                str(tmp_path),
                                self.generic_options)
        module.destroy()
        mock_action.assert_called_once()

    @patch('runway.cfngin.CFNgin.plan')
    def test_plan(self, mock_action, tmp_path):
        """Test plan."""
        module = CloudFormation(self.get_context(),
                                str(tmp_path),
                                self.generic_options)
        module.plan()
        mock_action.assert_called_once()



"""CFNgin test."""
# flake8: noqa
# pylint: disable=invalid-name
from os.path import basename

from integration_tests.test_cfngin.test_cfngin import Cfngin

FILE_BASENAME = '.'.join(basename(__file__).split('.')[:-1])


class TestSimpleBuild(Cfngin):
    """Test CFNgin simple build.

    Requires valid AWS credentials.

    """

    REQUIRED_FIXTURE_FILES = [
        FILE_BASENAME + '.yaml'
    ]
    TEST_NAME = __name__

    def _build(self):
        """Execute and assert initial build."""
        code, _stdout, stderr = self.runway_cmd('deploy')
        assert code == 0, 'exit code should be zero'
        expected_lines = [
            'Using default AWS provider mode',
            'simple-build-vpc: submitted (creating new stack)',
            'simple-build-vpc: complete (creating new stack)'
        ]
        for line in expected_lines:
            assert line in stderr, f'"{line}" missing from output'

    def _update_no_change(self):
        """Execute and assert second build with no changes."""
        code, _stdout, stderr = self.runway_cmd('deploy')
        assert code == 0, 'exit code should be zero'
        expected_lines = [
            'Using default AWS provider mode',
            'simple-build-vpc: skipped (nochange)'
        ]
        for line in expected_lines:
            assert line in stderr, f'"{line}" missing from output'

    def _destroy(self):
        """Execute and assert destroy."""
        code, _stdout, stderr = self.runway_cmd('destroy')
        assert code == 0, 'exit code should be zero'
        expected_lines = [
            'simple-build-vpc: submitted (submitted for destruction)',
            'simple-build-vpc: complete (stack destroyed)'
        ]
        for line in expected_lines:
            assert line in stderr, f'"{line}" missing from output'

    def run(self):
        """Run the test."""
        self.copy_fixtures()
        self._build()
        self._update_no_change()
        self._destroy()

    def teardown(self):
        """Teardown any created resources and delete files."""
        self.runway_cmd('destroy')  # cleanup incase of failure
        self.cleanup_fixtures()



"""Tests for module tags."""
from __future__ import print_function
import os
from copy import deepcopy

import boto3
import yaml
from send2trash import send2trash

from runway.util import change_dir

from integration_tests.integration_test import IntegrationTest
from integration_tests.util import (copy_dir, import_tests,
                                    execute_tests, run_command)

CFN_CLIENT = boto3.client('cloudformation', region_name='us-east-1')


class ModuleTags(IntegrationTest):
    """Base class for all module tag testing."""

    base_dir = os.path.abspath(os.path.dirname(__file__))
    tests_dir = os.path.join(base_dir, 'tests')
    sampleapp_dir = os.path.join(base_dir, 'sampleapp')

    stacker_file = {
        'namespace': 'runway-tests',
        'stacker_bucket': '',
        'stacks': {}
    }
    stack_definition = {
        'class_path': 'blueprints.fake_stack.BlueprintClass'
    }

    def check_stacks(self, should_exist):
        """Check to see if stacks exist.

        Args:
            should_exist (str): List of sampleapp stack numbers that should
                be deployed.

        """
        stacks = {}

        for i in range(1, 7):
            stack_name = 'runway-tests-module-tags-' + str(i)
            try:
                response = CFN_CLIENT.describe_stacks(
                    StackName=stack_name
                )['Stacks'][0]
                if response['StackStatus'] == 'CREATE_COMPLETE':
                    stacks[stack_name] = True
                else:
                    stacks[stack_name] = False
            except Exception:  # pylint: disable=broad-except
                stacks[stack_name] = False
        for stack, status in stacks.items():
            if stack[-1] in should_exist:
                assert status, stack + ' exists in the account'
            else:
                assert not status, stack + ' does not exist'
        return stacks

    def copy_sampleapp(self):
        """Create folder structure required for tests."""
        for i in range(1, 7):
            new_dir = os.path.join(self.base_dir, 'sampleapp' + str(i))
            copy_dir(os.path.join(self.base_dir, 'sampleapp'), new_dir)
            stacker_contents = deepcopy(self.stacker_file)
            stacker_contents['stacks'] = {
                'module-tags-' + str(i): self.stack_definition
            }
            with open(os.path.join(new_dir, 'stacker.yml'), 'w+') as yml:
                yml.write(yaml.safe_dump(stacker_contents))

    def runway_cmd(self, command, tags):
        """Run a deploy command based on tags."""
        cmd = ['runway', command]
        for tag in tags:
            cmd.append('--tag')
            cmd.append(tag)
        self.logger.info('Running command: %s', str(cmd))
        with change_dir(self.base_dir):
            return run_command(cmd)

    def clean(self):
        """Delete test resources."""
        with change_dir(self.base_dir):
            for i in range(1, 7):
                dir_to_del = os.path.join(self.base_dir,
                                          'sampleapp' + str(i))
                if os.path.isdir(dir_to_del):
                    send2trash(dir_to_del)

    def run(self):
        """Find all tests and run them."""
        import_tests(self.logger, self.tests_dir, 'test_*')
        self.set_environment('dev')
        self.set_env_var('CI', '1')
        tests = [test(self.logger) for test in ModuleTags.__subclasses__()]
        if not tests:
            raise Exception('No tests were found.')
        self.logger.debug('FOUND TESTS: %s', tests)
        err_count = execute_tests(tests, self.logger)
        assert err_count == 0  # assert that all subtests were successful
        return err_count

    def teardown(self):
        """Teardown resources create during init."""
        self.unset_env_var('CI')
        self.logger.debug('Teardown is defined in the submodules, not '
                          'the "ModuleTags" parent class.')



"""CFNgin session caching."""
import logging
import os
import warnings

import boto3

from .ui import ui

LOGGER = logging.getLogger(__name__)

DEFAULT_PROFILE = None
DEPRECATION_MSG = ('Use of "get_session" without providing credentials or a '
                   'profile has been deprecated and will raise an error after '
                   'the next major release. Please use the "get_session" '
                   'method of the context object instead.')
# A global credential cache that can be shared among boto3 sessions. This is
# inherently threadsafe thanks to the GIL:
# https://docs.python.org/3/glossary.html#term-global-interpreter-lock
CREDENTIAL_CACHE = {}


def get_session(region=None,
                profile=None,
                access_key=None,
                secret_key=None,
                session_token=None):
    """Create a thread-safe boto3 session.

    Args:
        region (Optional[str]): The region for the session.
        profile (Optional[str]): The profile for the session.
        access_key (Optional[str]): AWS Access Key ID.
        secret_key (Optional[str]): AWS secret Access Key.
        session_token (Optional[str]): AWS session token.

    Returns:
        :class:`boto3.session.Session`: A thread-safe boto3 session.

    """
    if profile:
        LOGGER.debug('Building session using profile "%s" in region "%s"',
                     profile, region or 'default')
    elif access_key:
        LOGGER.debug('Building session with Access Key "%s" in region "%s"',
                     access_key, region or 'default')
    elif os.environ.get('AWS_ACCESS_KEY_ID'):
        # TODO raise an error so we don't need to modify os.environ for cfngin
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        # TODO uncomment log message after we update all internal use
        # LOGGER.warning(DEPRECATION_MSG)

    session = boto3.Session(aws_access_key_id=access_key,
                            aws_secret_access_key=secret_key,
                            aws_session_token=session_token,
                            region_name=region,
                            profile_name=profile)
    cred_provider = session._session.get_component('credential_provider')
    provider = cred_provider.get_provider('assume-role')
    provider.cache = CREDENTIAL_CACHE
    provider._prompter = ui.getpass
    return session



"""Tests for the Source type object."""
import logging
import os
import unittest

from runway.sources.source import Source

LOGGER = logging.getLogger('runway')


class SourceTester(unittest.TestCase):
    """Tests for the Source type object."""

    def test_fetch_not_implemented(self):
        """#fetch: By default a not implemented error should be thrown."""
        source = Source({})
        self.assertRaises(NotImplementedError, source.fetch)

    def test_when_no_cache_dir_parameter_in_config(self):
        """The default when no cache_dir is passed in the config"""
        source = Source({})
        self.assertEqual(source.cache_dir, os.path.expanduser('~/.runway_cache'))

    def test_a_cache_directory_is_created(self):
        """Ensure a cache directory is created"""
        source = Source({})
        self.assertTrue(os.path.isdir(source.cache_dir))

    def test_directory_path_is_properly_sanitized(self):
        """Ensure that path values are sanitized for folder creation."""
        dir_path = Source.sanitize_directory_path('foo@bar/baz:bop')
        self.assertEqual(dir_path, 'foo_bar_baz_bop')



"""Test deploying stacks using tags."""
import os
from subprocess import check_output

from integration_tests.test_commands.test_commands import Commands


class TestRunPython(Commands):
    """Tests run-python subcommand."""

    TEST_NAME = __name__

    def init(self):
        """Initialize test."""
        pass  # pylint: disable=unnecessary-pass

    def run(self):
        """Run tests."""
        fixtures_path = os.path.join(os.path.dirname(os.path.dirname(__file__)),
                                     'fixtures')
        with open(os.path.join(fixtures_path,
                               'buildpipeline.json'), 'r') as stream:
            expected_template = stream.read()
        generated_template = check_output(
            ['runway',
             'run-python',
             os.path.join(fixtures_path, 'buildpipeline.py')]
        ).decode()

        assert generated_template == expected_template

    def teardown(self):
        """Teardown any created resources."""
        pass  # pylint: disable=unnecessary-pass



"""Launches or updates CloudFormation stacks based on the given config.

CFNgin is smart enough to figure out if anything (the template or parameters)
have changed for a given stack. If nothing has changed, CFNgin will correctly
skip executing anything against the stack.

"""
from ...actions import build
from .base import BaseCommand, cancel


class Build(BaseCommand):
    """Build subcommand."""

    name = "build"
    description = __doc__

    def add_arguments(self, parser):
        """Add arguments."""
        super(Build, self).add_arguments(parser)
        self._add_argument_max_parallel(parser)
        self._add_argument_tail(parser)
        parser.add_argument("-o", "--outline", action="store_true",
                            help="Print an outline of what steps will be "
                                 "taken to build the stacks")
        parser.add_argument("--force", action="append", default=[],
                            metavar="STACKNAME", type=str,
                            help="If a stackname is provided to --force, it "
                                 "will be updated, even if it is locked in "
                                 "the config.")
        parser.add_argument("--targets", "--stacks", action="append",
                            metavar="STACKNAME", type=str,
                            help="Only work on the stacks given, and their "
                                 "dependencies. Can be specified more than "
                                 "once. If not specified then CFNgin will "
                                 "work on all stacks in the config file.")
        parser.add_argument("-d", "--dump", action="store", type=str,
                            help="Dump the rendered Cloudformation templates "
                                 "to a directory")

    def run(self, options):
        """Run the command."""
        super(Build, self).run(options)
        action = build.Action(options.context,
                              provider_builder=options.provider_builder,
                              cancel=cancel())
        action.execute(concurrency=options.max_parallel,
                       outline=options.outline,
                       tail=options.tail,
                       dump=options.dump)

    def get_context_kwargs(self, options):
        """Return a dictionary of kwargs that will be used with the Context.

        This allows commands to pass in any specific arguments they define to
        the context.

        Args:
            options (:class:`argparse.Namespace`): arguments that have been
                passed via the command line

        Returns:
            Dict[str, Any]: Dictionary that will be passed to Context
            initializer as kwargs.

        """
        return {"stack_names": options.targets, "force_stacks": options.force}



"""CFNgin prehook responsible for creation of Lambda@Edge functions."""

import logging
import os
import re
import tempfile
from distutils.dir_util import copy_tree  # pylint: disable=import-error, no-name-in-module
from tempfile import mkstemp
from typing import Any, Dict, Optional  # pylint: disable=unused-import

from runway.cfngin.hooks import aws_lambda

from runway.cfngin.providers.base import BaseProvider  # pylint: disable=unused-import

# The functions associated with Auth@Edge
FUNCTIONS = [
    'check_auth',
    'refresh_auth',
    'parse_auth',
    'sign_out',
    'http_headers'
]


LOGGER = logging.getLogger(__name__)


def write(context,  # type: context.Context
          provider,  # type: BaseProvider
          **kwargs  # type: Optional[Dict[str, Any]]
         ):  # noqa: E124
    # type: (...) -> Dict[str, Any]
    """Writes/Uploads the configured lambdas for Auth@Edge.

    Lambda@Edge does not have the ability to allow Environment variables
    at the time of this writing. In order to configure our lambdas with
    dynamic variables we first will go through and update a "shared" template
    with all of the configuration elements and add that to a temporary
    folder along with each of the individual Lambda@Edge functions. This
    temporary folder is then used with the CFNgin awsLambda hook to build
    the functions.

    Args:
        context (cfngin.Context): The CFNgin context
        provider (cfngin.Provider): The CFNgin provider

    Keyword Args:
        client_id (str): The ID of the Cognito User Pool Client
        cookie_settings (dict): The settings for our customized cookies
        http_headers (dict): The additional headers added to our requests
        oauth_scopes (List[str]): The validation scopes for our OAuth requests
        redirect_path_auth_refresh (str): The URL path for authorization refresh
            redirect (Correlates to the refresh auth lambda)
        redirect_path_sign_in (str): The URL path to be redirected to after
            sign in (Correlates to the parse auth lambda)
        redirect_path_sign_out (str): The URL path to be redirected to after
            sign out (Correlates to the root to be asked to resignin)
        user_pool_id (str): The ID of the Cognito User Pool
    """
    cognito_domain = context.hook_data['aae_domain_updater'].get('domain')
    config = {
        'client_id': kwargs['client_id'],
        'cognito_auth_domain': cognito_domain,
        'cookie_settings': kwargs['cookie_settings'],
        'http_headers': kwargs['http_headers'],
        'oauth_scopes': kwargs['oauth_scopes'],
        'redirect_path_auth_refresh': kwargs['redirect_path_refresh'],
        'redirect_path_sign_in': kwargs['redirect_path_sign_in'],
        'redirect_path_sign_out': kwargs['redirect_path_sign_out'],
        'user_pool_id': context.hook_data['aae_user_pool_id_retriever']['id']
    }

    # Shared file that contains the method called for configuration data
    path = os.path.join(
        os.path.dirname(__file__),
        'templates/shared.py'
    )
    context_dict = {}

    with open(path) as file_:
        # Dynamically replace our configuration values
        # in the shared.py template file with actual
        # calculated values
        shared = re.sub(
            r'{.+?(})$',
            str(config),
            file_.read(),
            1,
            flags=re.DOTALL | re.MULTILINE
        )

        filedir, temppath = mkstemp()

        # Save the file to a temp path
        with open(temppath, 'w') as tmp:
            tmp.write(shared)
            config = temppath
        os.close(filedir)

        # Get all of the different Auth@Edge functions
        for handler in FUNCTIONS:
            # Create a temporary folder
            dirpath = tempfile.mkdtemp()

            # Copy the template code for the specific Lambda function
            # to the temporary folder
            copy_tree(
                os.path.join(
                    os.path.dirname(__file__),
                    'templates/%s' % handler
                ),
                dirpath
            )

            # Save our dynamic configuration shared file to the
            # temporary folder
            with open(config) as shared:
                raw = shared.read()
                filename = 'shared.py'
                with open('%s/%s' % (dirpath, filename), 'wb') as newfile:
                    newfile.write(raw.encode())

            # Upload our temporary folder to our S3 bucket for
            # Lambda use
            lamb = aws_lambda.upload_lambda_functions(
                context,
                provider,
                bucket=kwargs['bucket'],
                functions={
                    handler: {
                        'path': dirpath,
                    }
                }
            )

            # Add the lambda code reference to our context_dict
            context_dict.update(lamb)

    return context_dict



"""Split lookup."""
# pylint: disable=arguments-differ,unused-argument
from runway.lookups.handlers.base import LookupHandler

TYPE_NAME = "split"


class SplitLookup(LookupHandler):
    """Split lookup."""

    @classmethod
    def handle(cls, value, context=None, provider=None, **kwargs):
        """Split the supplied string on the given delimiter, providing a list.

        Args:
            value (str): Parameter(s) given to this lookup.
            context (:class:`runway.cfngin.context.Context`): Context instance.
            provider (:class:`runway.cfngin.providers.base.BaseProvider`):
                Provider instance.

        Returns:
            str: Looked up value.

        Format of value::

            <delimiter>::<value>

        Example:
            ::

                Subnets: ${split ,::subnet-1,subnet-2,subnet-3}

            Would result in the variable `Subnets` getting a list consisting
            of::

                ["subnet-1", "subnet-2", "subnet-3"]

            This is particularly useful when getting an output from another
            stack that contains a list. For example, the standard vpc blueprint
            outputs the list of Subnets it creates as a pair of Outputs
            (``PublicSubnets``, ``PrivateSubnets``) that are comma separated,
            so you could use this in your config::

                Subnets: ${split ,::${output vpc::PrivateSubnets}}

        """
        try:
            delimiter, text = value.split("::", 1)
        except ValueError:
            raise ValueError("Invalid value for split: %s. Must be in "
                             "<delimiter>::<text> format." % value)

        return text.split(delimiter)



"""
Runway Overview.

Usage:
  runway (test|preflight)
  runway (plan|taxi) [--tag <tag>...]
  runway (deploy|takeoff) [--tag <tag>...]
  runway (destroy|dismantle) [--tag <tag>...]
  runway init
  runway gen-sample [<samplename>]
  runway whichenv
  runway envvars
  runway run-aws <awscli-args>...
  runway run-python <filename>
  runway run-stacker <stacker-args>...
  runway tfenv (install|run) [<tfenv-args>...]
  runway kbenv (install|run) [<kbenv-args>...]
  runway -h | --help
  runway --version

Options:
  -h --help                         Show this screen.
  --version                         Show version.
  --tag <tag>...                    Select modules for processing by tag
                                    or tags. This option can be specified
                                    more than once to build a list of tags
                                    that are treated as "AND". (ex.
                                    "--tag <tag1> --tag <tag2>" would select
                                    all modules with BOTH tags).

Example:
  runway deploy --tag app:some-app --tag tier:web  # app:some-app AND tier:web
  runway destroy --tag app:some-app --tag tier:web  # app:some-app AND tier:web

Help:
  * Set the DEPLOY_ENVIRONMENT environment variable to set/override the
    autodetected environment. Autodetection is done from git branches in the
    form of ENV- (e.g. the dev environment is deployed from ENV-dev branch)
    falling back to name of the parent folder of the module.
  * All deploy commands (e.g. sls deploy, tf apply) will be run interactively
    unless the CI environment variable is set.

"""

import logging
import os
import sys

from docopt import docopt

from . import __version__ as version
from .cfngin.logger import ColorFormatter
from .commands.command_loader import find_command_class

# replicate stacker's colorized logs until we implement something better
COLOR_FORMAT = "%(levelname)s:%(name)s:\033[%(color)sm%(message)s\033[39m"
LOGGER = logging.getLogger('runway')
HDLR = logging.StreamHandler()
HDLR.setFormatter(ColorFormatter(
    COLOR_FORMAT if sys.stdout.isatty() else logging.BASIC_FORMAT
))


def fix_hyphen_commands(raw_cli_arguments):
    """Update options to match their module names with underscores."""
    for i in ['gen-sample', 'run-aws', 'run-python', 'run-stacker']:
        raw_cli_arguments[i.replace('-', '_')] = raw_cli_arguments[i]
        raw_cli_arguments.pop(i)
    return raw_cli_arguments


def main():
    """Provide main CLI entrypoint."""
    if os.environ.get('DEBUG'):
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO,
                            handlers=[HDLR])
        # botocore info is spammy
        logging.getLogger('botocore').setLevel(logging.ERROR)

    cli_arguments = fix_hyphen_commands(docopt(__doc__, version=version))

    # at least one of these must be enabled, i.e. the value is 'True'... but unfortunately
    #  `docopts` doesn't give you the hierarchy... so given 'gen-sample cfn', there are
    #  TWO enabled items in the list, 'gen-sample' and 'cfn'
    possible_commands = [command for command, enabled in cli_arguments.items() if enabled]

    command_class = find_command_class(possible_commands)
    if command_class:
        command_class(cli_arguments).execute()
    else:
        LOGGER.error("class not found for command '%s'", possible_commands)



"""Tests for runway.cfngin.lookups.handlers.rxref."""
import unittest

from mock import MagicMock

from runway.cfngin.config import Config
from runway.cfngin.context import Context
from runway.cfngin.lookups.handlers.rxref import RxrefLookup


class TestRxrefHandler(unittest.TestCase):
    """Tests for runway.cfngin.lookups.handlers.rxref.RxrefLookup."""

    def setUp(self):
        """Run before tests."""
        self.provider = MagicMock()
        self.context = Context(
            config=Config({"namespace": "ns"})
        )

    def test_rxref_handler(self):
        """Test rxref handler."""
        self.provider.get_output.return_value = "Test Output"

        value = RxrefLookup.handle("fully-qualified-stack-name::SomeOutput",
                                   provider=self.provider,
                                   context=self.context)
        self.assertEqual(value, "Test Output")

        args = self.provider.get_output.call_args
        self.assertEqual(args[0][0], "ns-fully-qualified-stack-name")
        self.assertEqual(args[0][1], "SomeOutput")



"""CFNgin test."""
# flake8: noqa
# pylint: disable=invalid-name
import os.path
from os.path import basename

from send2trash import send2trash

from integration_tests.test_cfngin.test_cfngin import Cfngin

FILE_BASENAME = '.'.join(basename(__file__).split('.')[:-1])


class TestRecreateFailed(Cfngin):
    """Test CFNgin recreation of a failed stack.

    Requires valid AWS credentials.

    This will take a bit of time to run (~5 min) due to needing to create
    a failed stack, allow it to rollback, and redeploy.

    """

    REQUIRED_FIXTURE_FILES = [FILE_BASENAME + '.1.yaml',
                              FILE_BASENAME + '.2.yaml']
    TEST_NAME = __name__

    def _deploy_bad(self):
        """Deploy failing config."""
        code, _stdout, stderr = self.runway_cmd('deploy')
        assert code != 0, 'exit code should be non-zero since one config failed'
        expected_lines = [
            'recreate-failed: submitted (creating new stack)',
            'recreate-failed: submitted (rolling back new stack)',
            'recreate-failed: failed (rolled back new stack)',
            'The following steps failed: recreate-failed'
        ]
        for line in expected_lines:
            assert line in stderr, f'"{line}" missing from output'

    def _deploy_good(self):
        """Deploy good config."""
        # get rid of the bad config so the good one can run
        send2trash(os.path.join(self.working_dir, FILE_BASENAME + '.1.yaml'))
        code, _stdout, stderr = self.runway_cmd('deploy')
        assert code == 0, 'exit code should be zero'
        expected_lines = [
            'recreate-failed: submitted (destroying stack for re-creation)',
            'recreate-failed: submitted (creating new stack)',
            'recreate-failed: complete (creating new stack)'
        ]
        for line in expected_lines:
            assert line in stderr, f'"{line}" missing from output'

    def run(self):
        """Run the test."""
        self.copy_fixtures()
        self._deploy_bad()
        self._deploy_good()

    def teardown(self):
        """Teardown any created resources and delete files."""
        self.runway_cmd('destroy')
        self.cleanup_fixtures()
